# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'options.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QDialog,
    QFrame,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTextEdit,
    QWidget,
)


class Ui_Options(object):
    def setupUi(self, Options):
        if not Options.objectName():
            Options.setObjectName("Options")
        Options.resize(575, 580)
        self.gridLayout = QGridLayout(Options)
        self.gridLayout.setObjectName("gridLayout")
        self.hlZScalingFactor = QHBoxLayout()
        self.hlZScalingFactor.setObjectName("hlZScalingFactor")
        self.pbZScalingFactorHelp = QPushButton(Options)
        self.pbZScalingFactorHelp.setObjectName("pbZScalingFactorHelp")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.pbZScalingFactorHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbZScalingFactorHelp.setSizePolicy(sizePolicy)
        self.pbZScalingFactorHelp.setMaximumSize(QSize(20, 16777215))
        self.pbZScalingFactorHelp.setFocusPolicy(Qt.NoFocus)

        self.hlZScalingFactor.addWidget(self.pbZScalingFactorHelp)

        self.lbZScalingFactor = QLabel(Options)
        self.lbZScalingFactor.setObjectName("lbZScalingFactor")

        self.hlZScalingFactor.addWidget(self.lbZScalingFactor)

        self.hsZScalingFactor = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlZScalingFactor.addItem(self.hsZScalingFactor)

        self.leZScalingFactor = QLineEdit(Options)
        self.leZScalingFactor.setObjectName("leZScalingFactor")
        sizePolicy1 = QSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(
            self.leZScalingFactor.sizePolicy().hasHeightForWidth()
        )
        self.leZScalingFactor.setSizePolicy(sizePolicy1)

        self.hlZScalingFactor.addWidget(self.leZScalingFactor)

        self.gridLayout.addLayout(self.hlZScalingFactor, 2, 0, 1, 1)

        self.hlPlotExportDPI = QHBoxLayout()
        self.hlPlotExportDPI.setObjectName("hlPlotExportDPI")
        self.pbPlotExportDPIHelp = QPushButton(Options)
        self.pbPlotExportDPIHelp.setObjectName("pbPlotExportDPIHelp")
        sizePolicy.setHeightForWidth(
            self.pbPlotExportDPIHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbPlotExportDPIHelp.setSizePolicy(sizePolicy)
        self.pbPlotExportDPIHelp.setMaximumSize(QSize(20, 16777215))
        self.pbPlotExportDPIHelp.setFocusPolicy(Qt.NoFocus)

        self.hlPlotExportDPI.addWidget(self.pbPlotExportDPIHelp)

        self.lbPlotExportDPI = QLabel(Options)
        self.lbPlotExportDPI.setObjectName("lbPlotExportDPI")

        self.hlPlotExportDPI.addWidget(self.lbPlotExportDPI)

        self.hsPlotExportDPI = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlPlotExportDPI.addItem(self.hsPlotExportDPI)

        self.lePlotExportDPI = QLineEdit(Options)
        self.lePlotExportDPI.setObjectName("lePlotExportDPI")
        sizePolicy1.setHeightForWidth(
            self.lePlotExportDPI.sizePolicy().hasHeightForWidth()
        )
        self.lePlotExportDPI.setSizePolicy(sizePolicy1)

        self.hlPlotExportDPI.addWidget(self.lePlotExportDPI)

        self.gridLayout.addLayout(self.hlPlotExportDPI, 16, 0, 1, 1)

        self.hlEFORange = QHBoxLayout()
        self.hlEFORange.setObjectName("hlEFORange")
        self.pbEFORangeHelp = QPushButton(Options)
        self.pbEFORangeHelp.setObjectName("pbEFORangeHelp")
        sizePolicy.setHeightForWidth(
            self.pbEFORangeHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbEFORangeHelp.setSizePolicy(sizePolicy)
        self.pbEFORangeHelp.setMaximumSize(QSize(20, 16777215))
        self.pbEFORangeHelp.setFocusPolicy(Qt.NoFocus)

        self.hlEFORange.addWidget(self.pbEFORangeHelp)

        self.lbEFORange = QLabel(Options)
        self.lbEFORange.setObjectName("lbEFORange")

        self.hlEFORange.addWidget(self.lbEFORange)

        self.hsEFORange = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlEFORange.addItem(self.hsEFORange)

        self.leEFORangeMin = QLineEdit(Options)
        self.leEFORangeMin.setObjectName("leEFORangeMin")
        sizePolicy1.setHeightForWidth(
            self.leEFORangeMin.sizePolicy().hasHeightForWidth()
        )
        self.leEFORangeMin.setSizePolicy(sizePolicy1)

        self.hlEFORange.addWidget(self.leEFORangeMin)

        self.leEFORangeMax = QLineEdit(Options)
        self.leEFORangeMax.setObjectName("leEFORangeMax")
        sizePolicy1.setHeightForWidth(
            self.leEFORangeMax.sizePolicy().hasHeightForWidth()
        )
        self.leEFORangeMax.setSizePolicy(sizePolicy1)

        self.hlEFORange.addWidget(self.leEFORangeMax)

        self.gridLayout.addLayout(self.hlEFORange, 5, 0, 1, 1)

        self.hlWeightAvgLocByECO = QHBoxLayout()
        self.hlWeightAvgLocByECO.setObjectName("hlWeightAvgLocByECO")
        self.pbWeightAvgLocByECOHelp = QPushButton(Options)
        self.pbWeightAvgLocByECOHelp.setObjectName("pbWeightAvgLocByECOHelp")
        sizePolicy.setHeightForWidth(
            self.pbWeightAvgLocByECOHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbWeightAvgLocByECOHelp.setSizePolicy(sizePolicy)
        self.pbWeightAvgLocByECOHelp.setMaximumSize(QSize(20, 16777215))
        self.pbWeightAvgLocByECOHelp.setFocusPolicy(Qt.NoFocus)

        self.hlWeightAvgLocByECO.addWidget(self.pbWeightAvgLocByECOHelp)

        self.cbWeightAvgLocByECO = QCheckBox(Options)
        self.cbWeightAvgLocByECO.setObjectName("cbWeightAvgLocByECO")

        self.hlWeightAvgLocByECO.addWidget(self.cbWeightAvgLocByECO)

        self.gridLayout.addLayout(self.hlWeightAvgLocByECO, 13, 0, 1, 1)

        self.teHelp = QTextEdit(Options)
        self.teHelp.setObjectName("teHelp")
        self.teHelp.setFocusPolicy(Qt.NoFocus)

        self.gridLayout.addWidget(self.teHelp, 24, 0, 1, 1)

        self.hlEFOBinSize = QHBoxLayout()
        self.hlEFOBinSize.setObjectName("hlEFOBinSize")
        self.pbEFOBinSizeHelp = QPushButton(Options)
        self.pbEFOBinSizeHelp.setObjectName("pbEFOBinSizeHelp")
        sizePolicy.setHeightForWidth(
            self.pbEFOBinSizeHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbEFOBinSizeHelp.setSizePolicy(sizePolicy)
        self.pbEFOBinSizeHelp.setMaximumSize(QSize(20, 16777215))
        self.pbEFOBinSizeHelp.setFocusPolicy(Qt.NoFocus)

        self.hlEFOBinSize.addWidget(self.pbEFOBinSizeHelp)

        self.lbEFOBinSize = QLabel(Options)
        self.lbEFOBinSize.setObjectName("lbEFOBinSize")

        self.hlEFOBinSize.addWidget(self.lbEFOBinSize)

        self.hsEFOBinSize = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlEFOBinSize.addItem(self.hsEFOBinSize)

        self.leEFOBinSize = QLineEdit(Options)
        self.leEFOBinSize.setObjectName("leEFOBinSize")
        sizePolicy1.setHeightForWidth(
            self.leEFOBinSize.sizePolicy().hasHeightForWidth()
        )
        self.leEFOBinSize.setSizePolicy(sizePolicy1)

        self.hlEFOBinSize.addWidget(self.leEFOBinSize)

        self.gridLayout.addLayout(self.hlEFOBinSize, 3, 0, 1, 1)

        self.lbInfo = QLabel(Options)
        self.lbInfo.setObjectName("lbInfo")
        font = QFont()
        font.setItalic(True)
        self.lbInfo.setFont(font)

        self.gridLayout.addWidget(self.lbInfo, 0, 0, 1, 1)

        self.hlMinNumTraces = QHBoxLayout()
        self.hlMinNumTraces.setObjectName("hlMinNumTraces")
        self.pbMinTIDNumHelp = QPushButton(Options)
        self.pbMinTIDNumHelp.setObjectName("pbMinTIDNumHelp")
        sizePolicy.setHeightForWidth(
            self.pbMinTIDNumHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbMinTIDNumHelp.setSizePolicy(sizePolicy)
        self.pbMinTIDNumHelp.setMaximumSize(QSize(20, 16777215))
        self.pbMinTIDNumHelp.setFocusPolicy(Qt.NoFocus)

        self.hlMinNumTraces.addWidget(self.pbMinTIDNumHelp)

        self.lbMinTIDNum = QLabel(Options)
        self.lbMinTIDNum.setObjectName("lbMinTIDNum")

        self.hlMinNumTraces.addWidget(self.lbMinTIDNum)

        self.hsMinTIDNum = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlMinNumTraces.addItem(self.hsMinTIDNum)

        self.leMinTIDNum = QLineEdit(Options)
        self.leMinTIDNum.setObjectName("leMinTIDNum")
        sizePolicy1.setHeightForWidth(self.leMinTIDNum.sizePolicy().hasHeightForWidth())
        self.leMinTIDNum.setSizePolicy(sizePolicy1)

        self.hlMinNumTraces.addWidget(self.leMinTIDNum)

        self.gridLayout.addLayout(self.hlMinNumTraces, 1, 0, 1, 1)

        self.lbInfoImmediate = QLabel(Options)
        self.lbInfoImmediate.setObjectName("lbInfoImmediate")
        self.lbInfoImmediate.setFont(font)

        self.gridLayout.addWidget(self.lbInfoImmediate, 11, 0, 1, 1)

        self.pbSetDefault = QPushButton(Options)
        self.pbSetDefault.setObjectName("pbSetDefault")

        self.gridLayout.addWidget(self.pbSetDefault, 25, 0, 1, 1)

        self.hlCFRRange = QHBoxLayout()
        self.hlCFRRange.setObjectName("hlCFRRange")
        self.pbCFRRangeHelp = QPushButton(Options)
        self.pbCFRRangeHelp.setObjectName("pbCFRRangeHelp")
        sizePolicy.setHeightForWidth(
            self.pbCFRRangeHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbCFRRangeHelp.setSizePolicy(sizePolicy)
        self.pbCFRRangeHelp.setMaximumSize(QSize(20, 16777215))
        self.pbCFRRangeHelp.setFocusPolicy(Qt.NoFocus)

        self.hlCFRRange.addWidget(self.pbCFRRangeHelp)

        self.lbCFRRange = QLabel(Options)
        self.lbCFRRange.setObjectName("lbCFRRange")

        self.hlCFRRange.addWidget(self.lbCFRRange)

        self.hsCFRRange = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlCFRRange.addItem(self.hsCFRRange)

        self.leCFRRangeMin = QLineEdit(Options)
        self.leCFRRangeMin.setObjectName("leCFRRangeMin")
        sizePolicy1.setHeightForWidth(
            self.leCFRRangeMin.sizePolicy().hasHeightForWidth()
        )
        self.leCFRRangeMin.setSizePolicy(sizePolicy1)

        self.hlCFRRange.addWidget(self.leCFRRangeMin)

        self.leCFRRangeMax = QLineEdit(Options)
        self.leCFRRangeMax.setObjectName("leCFRRangeMax")
        sizePolicy1.setHeightForWidth(
            self.leCFRRangeMax.sizePolicy().hasHeightForWidth()
        )
        self.leCFRRangeMax.setSizePolicy(sizePolicy1)

        self.hlCFRRange.addWidget(self.leCFRRangeMax)

        self.gridLayout.addLayout(self.hlCFRRange, 6, 0, 1, 1)

        self.lbInfoAtStart = QLabel(Options)
        self.lbInfoAtStart.setObjectName("lbInfoAtStart")
        self.lbInfoAtStart.setFont(font)

        self.gridLayout.addWidget(self.lbInfoAtStart, 19, 0, 1, 1)

        self.lbQuickHelp = QLabel(Options)
        self.lbQuickHelp.setObjectName("lbQuickHelp")
        font1 = QFont()
        font1.setBold(True)
        self.lbQuickHelp.setFont(font1)

        self.gridLayout.addWidget(self.lbQuickHelp, 22, 0, 1, 1)

        self.hl_at_start = QFrame(Options)
        self.hl_at_start.setObjectName("hl_at_start")
        self.hl_at_start.setFrameShape(QFrame.HLine)
        self.hl_at_start.setFrameShadow(QFrame.Sunken)

        self.gridLayout.addWidget(self.hl_at_start, 18, 0, 1, 1)

        self.hlLocPrecRange = QHBoxLayout()
        self.hlLocPrecRange.setObjectName("hlLocPrecRange")
        self.pbLocPrecRangeHelp = QPushButton(Options)
        self.pbLocPrecRangeHelp.setObjectName("pbLocPrecRangeHelp")
        sizePolicy.setHeightForWidth(
            self.pbLocPrecRangeHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbLocPrecRangeHelp.setSizePolicy(sizePolicy)
        self.pbLocPrecRangeHelp.setMaximumSize(QSize(20, 16777215))
        self.pbLocPrecRangeHelp.setFocusPolicy(Qt.NoFocus)

        self.hlLocPrecRange.addWidget(self.pbLocPrecRangeHelp)

        self.lbLocPrecRange = QLabel(Options)
        self.lbLocPrecRange.setObjectName("lbLocPrecRange")

        self.hlLocPrecRange.addWidget(self.lbLocPrecRange)

        self.hsLocPrecRange = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlLocPrecRange.addItem(self.hsLocPrecRange)

        self.leLocPrecRangeMin = QLineEdit(Options)
        self.leLocPrecRangeMin.setObjectName("leLocPrecRangeMin")
        sizePolicy1.setHeightForWidth(
            self.leLocPrecRangeMin.sizePolicy().hasHeightForWidth()
        )
        self.leLocPrecRangeMin.setSizePolicy(sizePolicy1)

        self.hlLocPrecRange.addWidget(self.leLocPrecRangeMin)

        self.leLocPrecRangeMax = QLineEdit(Options)
        self.leLocPrecRangeMax.setObjectName("leLocPrecRangeMax")
        sizePolicy1.setHeightForWidth(
            self.leLocPrecRangeMax.sizePolicy().hasHeightForWidth()
        )
        self.leLocPrecRangeMax.setSizePolicy(sizePolicy1)

        self.hlLocPrecRange.addWidget(self.leLocPrecRangeMax)

        self.gridLayout.addLayout(self.hlLocPrecRange, 7, 0, 1, 1)

        self.hl_immediate = QFrame(Options)
        self.hl_immediate.setObjectName("hl_immediate")
        self.hl_immediate.setFrameShape(QFrame.HLine)
        self.hl_immediate.setFrameShadow(QFrame.Sunken)

        self.gridLayout.addWidget(self.hl_immediate, 10, 0, 1, 1)

        self.hlExpectedEFOFrequency = QHBoxLayout()
        self.hlExpectedEFOFrequency.setObjectName("hlExpectedEFOFrequency")
        self.pbEFOSingleEmitterFrequencyHelp = QPushButton(Options)
        self.pbEFOSingleEmitterFrequencyHelp.setObjectName(
            "pbEFOSingleEmitterFrequencyHelp"
        )
        sizePolicy.setHeightForWidth(
            self.pbEFOSingleEmitterFrequencyHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbEFOSingleEmitterFrequencyHelp.setSizePolicy(sizePolicy)
        self.pbEFOSingleEmitterFrequencyHelp.setMaximumSize(QSize(20, 16777215))
        self.pbEFOSingleEmitterFrequencyHelp.setFocusPolicy(Qt.NoFocus)

        self.hlExpectedEFOFrequency.addWidget(self.pbEFOSingleEmitterFrequencyHelp)

        self.lbEFOSingleEmitterFrequency = QLabel(Options)
        self.lbEFOSingleEmitterFrequency.setObjectName("lbEFOSingleEmitterFrequency")

        self.hlExpectedEFOFrequency.addWidget(self.lbEFOSingleEmitterFrequency)

        self.hsEFOSingleEmitterFrequency = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlExpectedEFOFrequency.addItem(self.hsEFOSingleEmitterFrequency)

        self.leEFOSingleEmitterFrequency = QLineEdit(Options)
        self.leEFOSingleEmitterFrequency.setObjectName("leEFOSingleEmitterFrequency")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.leEFOSingleEmitterFrequency.sizePolicy().hasHeightForWidth()
        )
        self.leEFOSingleEmitterFrequency.setSizePolicy(sizePolicy2)

        self.hlExpectedEFOFrequency.addWidget(self.leEFOSingleEmitterFrequency)

        self.gridLayout.addLayout(self.hlExpectedEFOFrequency, 4, 0, 1, 1)

        self.hlShowConsoleAtStart = QHBoxLayout()
        self.hlShowConsoleAtStart.setObjectName("hlShowConsoleAtStart")
        self.pbOpenConsoleAtStartHelp = QPushButton(Options)
        self.pbOpenConsoleAtStartHelp.setObjectName("pbOpenConsoleAtStartHelp")
        sizePolicy.setHeightForWidth(
            self.pbOpenConsoleAtStartHelp.sizePolicy().hasHeightForWidth()
        )
        self.pbOpenConsoleAtStartHelp.setSizePolicy(sizePolicy)
        self.pbOpenConsoleAtStartHelp.setMaximumSize(QSize(20, 16777215))
        self.pbOpenConsoleAtStartHelp.setFocusPolicy(Qt.NoFocus)

        self.hlShowConsoleAtStart.addWidget(self.pbOpenConsoleAtStartHelp)

        self.cbOpenConsoleAtStart = QCheckBox(Options)
        self.cbOpenConsoleAtStart.setObjectName("cbOpenConsoleAtStart")

        self.hlShowConsoleAtStart.addWidget(self.cbOpenConsoleAtStart)

        self.gridLayout.addLayout(self.hlShowConsoleAtStart, 20, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(
            20, 10, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout.addItem(self.verticalSpacer, 21, 0, 1, 1)

        QWidget.setTabOrder(self.leMinTIDNum, self.leZScalingFactor)
        QWidget.setTabOrder(self.leZScalingFactor, self.leEFOBinSize)
        QWidget.setTabOrder(self.leEFOBinSize, self.leEFOSingleEmitterFrequency)
        QWidget.setTabOrder(self.leEFOSingleEmitterFrequency, self.leEFORangeMin)
        QWidget.setTabOrder(self.leEFORangeMin, self.leEFORangeMax)
        QWidget.setTabOrder(self.leEFORangeMax, self.leCFRRangeMin)
        QWidget.setTabOrder(self.leCFRRangeMin, self.leCFRRangeMax)
        QWidget.setTabOrder(self.leCFRRangeMax, self.leLocPrecRangeMin)
        QWidget.setTabOrder(self.leLocPrecRangeMin, self.leLocPrecRangeMax)
        QWidget.setTabOrder(self.leLocPrecRangeMax, self.cbWeightAvgLocByECO)
        QWidget.setTabOrder(self.cbWeightAvgLocByECO, self.lePlotExportDPI)
        QWidget.setTabOrder(self.lePlotExportDPI, self.cbOpenConsoleAtStart)
        QWidget.setTabOrder(self.cbOpenConsoleAtStart, self.pbSetDefault)

        self.retranslateUi(Options)

        QMetaObject.connectSlotsByName(Options)

    # setupUi

    def retranslateUi(self, Options):
        Options.setWindowTitle(QCoreApplication.translate("Options", "Options", None))
        self.pbZScalingFactorHelp.setText(
            QCoreApplication.translate("Options", "?", None)
        )
        # if QT_CONFIG(tooltip)
        self.lbZScalingFactor.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbZScalingFactor.setText(
            QCoreApplication.translate("Options", "Z scaling factor", None)
        )
        # if QT_CONFIG(tooltip)
        self.leZScalingFactor.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.pbPlotExportDPIHelp.setText(
            QCoreApplication.translate("Options", "?", None)
        )
        self.lbPlotExportDPI.setText(
            QCoreApplication.translate(
                "Options", "Resolution (DPI) for exporting plots", None
            )
        )
        self.pbEFORangeHelp.setText(QCoreApplication.translate("Options", "?", None))
        # if QT_CONFIG(tooltip)
        self.lbEFORange.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbEFORange.setText(
            QCoreApplication.translate("Options", "EFO default plot range", None)
        )
        # if QT_CONFIG(tooltip)
        self.leEFORangeMin.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(tooltip)
        self.leEFORangeMax.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.pbWeightAvgLocByECOHelp.setText(
            QCoreApplication.translate("Options", "?", None)
        )
        # if QT_CONFIG(tooltip)
        self.cbWeightAvgLocByECO.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.cbWeightAvgLocByECO.setText(
            QCoreApplication.translate(
                "Options",
                "Use relative ECO count for weighted average localization calculation",
                None,
            )
        )
        self.pbEFOBinSizeHelp.setText(QCoreApplication.translate("Options", "?", None))
        # if QT_CONFIG(tooltip)
        self.lbEFOBinSize.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbEFOBinSize.setText(
            QCoreApplication.translate("Options", "EFO bin size (Hz)", None)
        )
        # if QT_CONFIG(tooltip)
        self.leEFOBinSize.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbInfo.setText(
            QCoreApplication.translate(
                "Options",
                "Changes below will be applied when loading new data or when applying filters.",
                None,
            )
        )
        self.pbMinTIDNumHelp.setText(QCoreApplication.translate("Options", "?", None))
        # if QT_CONFIG(tooltip)
        self.lbMinTIDNum.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbMinTIDNum.setText(
            QCoreApplication.translate(
                "Options", "Minimum number of trace localizations", None
            )
        )
        # if QT_CONFIG(tooltip)
        self.leMinTIDNum.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbInfoImmediate.setText(
            QCoreApplication.translate(
                "Options",
                "Changes below will be applied immediately to all open views.",
                None,
            )
        )
        self.pbSetDefault.setText(
            QCoreApplication.translate("Options", "Set as new default", None)
        )
        self.pbCFRRangeHelp.setText(QCoreApplication.translate("Options", "?", None))
        # if QT_CONFIG(tooltip)
        self.lbCFRRange.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbCFRRange.setText(
            QCoreApplication.translate("Options", "CFR default plot range", None)
        )
        # if QT_CONFIG(tooltip)
        self.leCFRRangeMin.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(tooltip)
        self.leCFRRangeMax.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbInfoAtStart.setText(
            QCoreApplication.translate(
                "Options", "Changes below will be applied at application start.", None
            )
        )
        self.lbQuickHelp.setText(
            QCoreApplication.translate("Options", "Quick help", None)
        )
        self.pbLocPrecRangeHelp.setText(
            QCoreApplication.translate("Options", "?", None)
        )
        # if QT_CONFIG(tooltip)
        self.lbLocPrecRange.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbLocPrecRange.setText(
            QCoreApplication.translate(
                "Options", "Localization precision default plot range", None
            )
        )
        # if QT_CONFIG(tooltip)
        self.leLocPrecRangeMin.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        # if QT_CONFIG(tooltip)
        self.leLocPrecRangeMax.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.pbEFOSingleEmitterFrequencyHelp.setText(
            QCoreApplication.translate("Options", "?", None)
        )
        # if QT_CONFIG(tooltip)
        self.lbEFOSingleEmitterFrequency.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.lbEFOSingleEmitterFrequency.setText(
            QCoreApplication.translate(
                "Options", "EFO expected frequency for single emitters (Hz)", None
            )
        )
        # if QT_CONFIG(tooltip)
        self.leEFOSingleEmitterFrequency.setToolTip("")
        # endif // QT_CONFIG(tooltip)
        self.pbOpenConsoleAtStartHelp.setText(
            QCoreApplication.translate("Options", "?", None)
        )
        self.cbOpenConsoleAtStart.setText(
            QCoreApplication.translate("Options", "Show console at start", None)
        )

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import math
from typing import Optional

import numpy as np
from scipy import stats
from scipy.ndimage import median_filter
from scipy.signal import find_peaks
from sklearn.mixture import BayesianGaussianMixture


def hist_bins(values: np.ndarray, bin_size: float) -> tuple:
    """Return the bins to be used for the passed values and bin size.

    Parameters
    ----------

    values: np.ndarray
        One-dimensional array of values for which to determine the ideal histogram bins.

    bin_size: float
        Bin size to use.

    Returns
    -------

    bin_edges: np.ndarray
        Array of bin edges (to use with np.histogram()).

    bin_centers: np.ndarray
        Array of bin centers.

    bin_width:
        Bin width.
    """

    if len(values) == 0:
        raise ValueError("No data.")

    # Find an appropriate min value that keeps the bins nicely centered
    min_value = bin_size * int(np.min(values) / bin_size)

    # Max value
    max_value = np.max(values)

    # Pathological case where bin_width is 0.0
    if bin_size <= 0.0:
        raise ValueError("`bin_size` must be a positive number!")

    # Calculate number of bins
    num_bins = math.floor((max_value - min_value) / bin_size) + 1

    # Center the first bin around the min value
    half_width = bin_size / 2
    bin_edges = np.arange(
        min_value - half_width, min_value + num_bins * bin_size, bin_size
    )
    bin_centers = (bin_edges[0:-1] + bin_edges[1:]) / 2

    return bin_edges, bin_centers, bin_size


def ideal_hist_bins(values: np.ndarray, scott: bool = False):
    """Calculate the ideal histogram bins using the Freedman-Diaconis rule.

    See: https://en.wikipedia.org/wiki/Freedman%E2%80%93Diaconis_rule

    Parameters
    ----------

    values: np.ndarray
        One-dimensional array of values for which to determine the ideal histogram bins.

    scott: bool
        Whether to use Scott's normal reference rule (if the data is normally distributed).

    Returns
    -------

    bin_edges: np.ndarray
        Array of bin edges (to use with np.histogram()).

    bin_centers: np.ndarray
        Array of bin centers.

    bin_size:
        Bin width.
    """

    if len(values) == 0:
        raise ValueError("No data.")

    # Pathological case, all values are the same
    if np.all(np.diff(values) == 0):
        bin_edges = (values[0] - 5e-7, values[0] + 5e-7)
        bin_centers = (values[0],)
        bin_size = 1e-6
        return bin_edges, bin_centers, bin_size

    # Calculate bin width
    factor = 2.0
    if scott:
        factor = 2.59
    iqr = stats.iqr(values, rng=(25, 75), scale=1.0, nan_policy="omit")
    num_values = np.sum(np.logical_not(np.isnan(values)))
    crn = np.power(num_values, 1 / 3)
    bin_size = (factor * iqr) / crn

    # Get min and max values
    min_value = np.min(values)
    max_value = np.max(values)

    # Pathological case where bin_size is 0.0
    if bin_size == 0.0:
        bin_size = 0.5 * (min_value + max_value)

    # Calculate number of bins
    num_bins = math.floor((max_value - min_value) / bin_size) + 1

    # Center the first bin around the min value
    half_width = bin_size / 2
    bin_edges = np.arange(
        min_value - half_width, min_value + num_bins * bin_size, bin_size
    )
    bin_centers = (bin_edges[0:-1] + bin_edges[1:]) / 2
    if len(bin_edges) >= 2:
        bin_size = bin_edges[1] - bin_edges[0]

    return bin_edges, bin_centers, bin_size


def get_robust_threshold(values: np.ndarray, factor: float = 2.0):
    """Calculate a robust threshold for the array of values.

    The threshold is defines as `median + thresh * median absolute deviation`.

    The median absolute deviation is divided by 0.67449 to bring it in the
    same scale as the (non-robust) standard deviation.

    Parameters
    ----------

    values: np.ndarray
        Array of values. It may contain NaNs.

    factor: float
        Factor by which to multiply the median absolute deviation.

    Returns
    -------

    upper_threshold: float
        Upper threshold.

    lower_threshold: float
        Lower threshold.

    med: float
        Median of the array of values.

    mad: float
        Scaled median absolute deviation of the array of values.
    """

    # Remove NaNs
    work_values = values.copy()
    work_values = work_values[np.logical_not(np.isnan(work_values))]
    if len(work_values) == 0:
        return None, None, None, None

    # Calculate robust statistics and threshold
    med = np.median(work_values)
    mad = stats.median_abs_deviation(work_values, scale=0.67449)
    step = factor * mad
    upper_threshold = med + step
    lower_threshold = med - step

    return upper_threshold, lower_threshold, med, mad


def prepare_histogram(
    values: np.ndarray,
    normalize: bool = True,
    auto_bins: bool = True,
    scott: bool = False,
    bin_size: float = 0.0,
):
    """Return histogram counts and bins for given values with provided or automatically calculated bin number.

    Parameters
    ----------

    values: np.ndarray
        Array of values. It may contain NaNs.

    normalize: bool
        Whether to normalize the histogram to a probability mass function (PMF). The integral of the PMF is 1.0.

    auto_bins: bool
        Whether to automatically calculate the bin size from the data.

    scott: bool
        Whether to use Scott's normal reference rule (the data should be normally distributed). This is used only
        if `auto_bins` is True.

    bin_size: float
        Bin size to use if `auto_bins` is False. It will be ignored if `auto_bins` is True.

    Returns
    -------

    n: np.ndarray
        Histogram counts (optionally normalized to sum to 1.0).

    bin_edges: np.ndarray
        Array of bin edges (to use with np.histogram()).

    bin_centers: np.ndarray
        Array of bin centers.

    bin_width:
        Bin width.

    """
    if auto_bins:
        bin_edges, bin_centers, bin_width = ideal_hist_bins(values, scott=scott)
    else:
        if bin_size == 0.0:
            raise Exception(
                f"Please provide a valid value for `bin_size` if `auto_bins` is False."
            )
        bin_edges, bin_centers, bin_width = hist_bins(values, bin_size=bin_size)

    n, _ = np.histogram(values, bins=bin_edges, density=False)
    if normalize:
        n = n / n.sum()
    return n, bin_edges, bin_centers, bin_width


def find_first_peak_bounds(
    counts: np.ndarray,
    bins: np.ndarray,
    min_rel_prominence: float = 0.01,
    med_filter_support: int = 5,
):
    """Finds the first peak in the histogram and return the lower and upper bounds.

    Parameters
    ----------

    counts: np.ndarray
        Array of histogram counts.

    bins: np.ndarray
        Array of histogram bins.

    min_rel_prominence: float
        Minimum relative prominences (relative to range of filtered counts) for peaks to be considered valid.

    med_filter_support: int
        Support for the median filter to suppress some spurious noisy peaks in the counts.

    Returns
    -------

    lower_bound: float
        Lower bound of the first peak.

    upper_bound: float
        Upper bound of the first peak.
    """

    # Filter the signal
    x = median_filter(counts, footprint=np.ones(med_filter_support))

    # Absolute minimum prominence
    min_prominence = min_rel_prominence * (x.max() - x.min())

    # Find maxima
    peaks, properties = find_peaks(x, prominence=(min_prominence, None))

    # Find minima
    x_inv = x.max() - x
    peaks_inv, properties_inv = find_peaks(x_inv, prominence=(min_prominence, None))

    # If we did not find any local maxima, we return failure
    if len(peaks) == 0:
        return None, None

    # First peak position
    first_peak = peaks[0]

    # If we do not have any local minima, we return the beginning and end of the bins range
    if len(peaks_inv) == 0:
        return bins[0], bins[-1]

    # Do we have a minimum on the left of the first peak?
    candidates_left = peaks_inv[peaks_inv < first_peak]
    if len(candidates_left) == 0:
        lower_bound = bins[0]
    else:
        lower_bound = bins[candidates_left[-1]]

    # Do we have a minimum on the right of the first peak?
    candidates_right = peaks_inv[peaks_inv > first_peak]
    if len(candidates_right) == 0:
        upper_bound = bins[-1]
    else:
        upper_bound = bins[candidates_right[0]]

    return lower_bound, upper_bound


def find_cutoff_near_value(
    counts: np.ndarray,
    bins: np.ndarray,
    expected_value: float,
):
    """Finds the first peak in the histogram and return the lower and upper bounds.

    Parameters
    ----------

    counts: np.ndarray
        Array of histogram counts.

    bins: np.ndarray
        Array of histogram bins.

    expected_value: float
        The cutoff is expected to be close to the expected value.

    Returns
    -------

    cutoff: float
        Estiated cutoff frequency.
    """

    # Absolute minimum prominence
    min_prominence = 0.05 * (counts.max() - counts.min())

    # Find minima
    counts_inv = counts.max() - counts
    peaks_inv, properties_inv = find_peaks(
        counts_inv, prominence=(min_prominence, None)
    )

    # Which is the local minimum closest to the expected value
    cutoff_pos = peaks_inv[np.argmin(np.abs(bins[peaks_inv] - expected_value))]

    # Extract the corresponding frequency
    cutoff = bins[cutoff_pos]

    # Return the obtained cutoff frequency
    return cutoff


def calculate_density_map(
    x: np.ndarray,
    y: np.ndarray,
    x_bin_edges: Optional[np.ndarray] = None,
    y_bin_edges: Optional[np.ndarray] = None,
    auto_bins: bool = True,
    scott: bool = False,
    bin_size: Optional[float] = None,
) -> np.ndarray:
    """Create density map for 2D data.

    Parameters
    ----------

    x: np.ndarray
        1D array of X values.

    y: np.ndarray
        1D array of Y values.

    x_bin_edges: np.ndarray
        1D array of bin edge values for the X array. If omitted, it will be calculated automatically
        (see `pyminflux.analysis.prepare_histogram`.)

    y_bin_edges: np.ndarray
        1D array of bin edge values for the X array. If omitted, it will be calculated automatically
        (see `pyminflux.analysis.prepare_histogram`.)

    auto_bins: bool
        Whether to automatically calculate the bin size from the data. Only used if either `x_bin_edges`
        or `y_bin_edges` are None.

    scott: bool
        Whether to use Scott's normal reference rule (the data should be normally distributed).
        This is only used if either `x_bin_edges` or `y_bin_edges` are None and `auto_bins` is True.

    bin_size: float
        Bin size to use if either `x_bin_edges` or `y_bin_edges` are None and `auto_bins` is False.
        It will be ignored if `auto_bins` is True.

    Returns
    -------

    density: np.ndarray
        2D density maps.
    """

    # Calculate bin edges if needed
    if x_bin_edges is None:
        _, x_bin_edges, _, _ = prepare_histogram(
            x, auto_bins=auto_bins, scott=scott, bin_size=bin_size
        )

    if y_bin_edges is None:
        _, y_bin_edges, _, _ = prepare_histogram(
            y, auto_bins=auto_bins, scott=scott, bin_size=bin_size
        )

    # Create density map
    xx, yy = np.meshgrid(x_bin_edges, y_bin_edges)
    positions = np.vstack([xx.ravel(), yy.ravel()])
    values = np.vstack([x, y])
    kernel = stats.gaussian_kde(values)
    density = np.reshape(kernel(positions).T, xx.shape)

    # Return density map
    return density


def calculate_2d_histogram(
    x: np.ndarray,
    y: np.ndarray,
    x_bin_edges: Optional[np.ndarray] = None,
    y_bin_edges: Optional[np.ndarray] = None,
    x_auto_bins: bool = True,
    y_auto_bins: bool = True,
    scott: bool = False,
    x_bin_size: float = 0.0,
    y_bin_size: float = 0.0,
) -> np.ndarray:
    """Create density map for 2D data.

    Parameters
    ----------

    x: np.ndarray
        1D array of X values.

    y: np.ndarray
        1D array of Y values.

    x_bin_edges: np.ndarray
        1D array of bin edge values for the X array. If omitted, it will be calculated automatically
        (see `pyminflux.analysis.prepare_histogram`.)

    y_bin_edges: np.ndarray
        1D array of bin edge values for the X array. If omitted, it will be calculated automatically
        (see `pyminflux.analysis.prepare_histogram`.)

    x_auto_bins: bool
        Whether to automatically calculate the bin size for `x` from the data. Only used if `x_bin_edges`
        is None.

    y_auto_bins: bool
        Whether to automatically calculate the bin size for `y` from the data. Only used if `y_bin_edges`
        is None.

    scott: bool
        Whether to use Scott's normal reference rule (the data should be normally distributed).
        This is only used if either `x_bin_edges` or `y_bin_edges` are None and `auto_bins` is True.

    x_bin_size: float
        Bin size to use for `x` if `x_bin_edges` is None and `x_auto_bins` is False.
        It will be ignored if `x_auto_bins` is True.

    y_bin_size: float
        Bin size to use for `y` if `y_bin_edges` is None and `y_auto_bins` is False.
        It will be ignored if `y_auto_bins` is True.

    Returns
    -------

    density: np.ndarray
        2D density maps.
    """

    # Calculate bin edges if needed
    if x_bin_edges is None:
        _, x_bin_edges, _, _ = prepare_histogram(
            x, auto_bins=x_auto_bins, scott=scott, bin_size=x_bin_size
        )

    if y_bin_edges is None:
        _, y_bin_edges, _, _ = prepare_histogram(
            y, auto_bins=y_auto_bins, scott=scott, bin_size=y_bin_size
        )

    # Create 2D histogram
    histogram = np.histogram2d(y, x, bins=(y_bin_edges, x_bin_edges))

    # Return histogram
    return histogram[0]


def assign_data_to_clusters(
    x: np.ndarray, num_clusters: int = 2, seed: Optional[int] = None
):
    """Use Gaussian Mixture Model fitting to assign data to the requested number of clusters.

    Parameters
    ----------

    x: np.ndarray
        Array of values to cluster.

    num_clusters: int (Default = 2)
        Number of clusters to be assigned.

    seed: int (Optional)
        Seed for the random number generator.

    Returns
    -------

    ids: np.ndarray
        Array of cluster ids for each value in x. The first cluster has ID 1.
    """

    # Make sure to work with a NumPy array
    x = np.array(x)

    # Make sure to work with a column vector
    x = x.reshape(-1, 1)

    # Fit a Bayesian Gaussian Mixture Model with self.state.num_fluorophores components
    model = BayesianGaussianMixture(
        n_components=num_clusters,
        init_params="k-means++",
        covariance_type="full",
        max_iter=1000,
        random_state=seed,
    ).fit(x)

    # Predict with the selected model
    y_pred = model.predict(x)

    # If num_clusters is > 1, sort the indices by mean values of x,
    # from low to high (to match the assignment of the manual thresholding)
    if num_clusters > 1:
        means = [np.mean(x[y_pred == f_id]) for f_id in np.unique(y_pred)]
        sorted_f = np.argsort(means)
        for f in range(num_clusters):
            if np.isnan(means[f]):
                continue
            n = sorted_f[f] + num_clusters
            y_pred[y_pred == f] = n
        y_pred -= num_clusters

    # Now make the clusters ID start from 1 instead of 0.
    ids = y_pred + 1

    # Return
    return ids


def reassign_fluo_ids_by_majority_vote(fluo_ids: np.ndarray, tids: np.ndarray):
    """Reassign IDs of fluorophores by majority so that TIDs have only one fluorophore ID assigned."""

    # Work on a copy of fluo_ids
    work_fluo_ids = fluo_ids.copy()

    # Find the unique tids and their indices
    _, indices = np.unique(tids, return_index=True)

    # Split the fluo_ids array into sub-arrays corresponding to each unique ID.
    # Note: np.split() returns views into the original array, so the code below
    # will update work_fluo_ids in place!
    split_fluo_ids = np.split(work_fluo_ids, indices[1:])

    # Reassign by majority vote
    for i, s in enumerate(split_fluo_ids):
        if not np.all(s == s[0]):
            # Only process tids that have more than one fluo_id associated to them
            majority_class = np.argmax(np.bincount(s))
            split_fluo_ids[i][:] = majority_class

    # We can now return work_fluo_ids, that has been modified in place by the split_fluo_ids views
    return work_fluo_ids



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import re

import requests

import pyminflux


def check_for_updates():
    """Check for pyMINFLUX updates.

    Returns
    -------

    code: int
        Success code:
        -1: something went wrong retrieving version information.
         0: there are no new versions.
         1: there is a new version.
    version: str
        Version on the server (in the format x.y.z). Set if code is 0 or 1.
    error: str
        Error message. Only set if code is -1.
    """

    # Initialize outputs
    code = -1
    version = ""
    error = ""

    # Get the redirect from the latest release URL
    response = requests.get(
        "https://github.com/bsse-scf/pyMINFLUX/releases/latest", allow_redirects=False
    )

    # This should redirect (status code 301 or 302)
    if response.status_code in (301, 302):
        redirect_url = response.headers["Location"]
    else:
        error = "Could not check for updates!"
        return code, version, error

    # Try retrieving the version string
    match = re.search(r"\b(\d+)\.(\d+)\.(\d+)$", redirect_url)
    if match:
        x, y, z = match.groups()
    else:
        error = "Could not retrieve version information from server!"
        return code, version, error

    # Set the version
    version = f"{x}.{y}.{z}"

    # Transform new version into an integer
    new_version = 10000 * int(x) + 100 * int(y) + int(z)

    # Current version
    parts = pyminflux.__version__.split(".")

    # Make sure that we have three parts
    if len(parts) != 3:
        error = "Could not retrieve current app information!"
        return code, version, error

    # Transform current version into an integer
    current_version = 10000 * int(parts[0]) + 100 * int(parts[1]) + int(parts[2])

    # Now check
    if new_version > current_version:
        code = 1
    else:
        code = 0

    # Return
    return code, version, error


def intersect_2d_ranges(first_range, second_range):
    """Intersect two 1D ranges (min, max) to get the combined results of two consecutive filtering events."""
    out_range = (
        max(first_range[0], second_range[0]),
        min(first_range[1], second_range[1]),
    )
    return out_range



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#


__doc__ = "Analysis functions."
__all__ = [
    "assign_data_to_clusters",
    "calculate_2d_histogram",
    "calculate_density_map",
    "find_cutoff_near_value",
    "find_first_peak_bounds",
    "get_robust_threshold",
    "ideal_hist_bins",
    "prepare_histogram",
    "reassign_fluo_ids_by_majority_vote",
]

from ._analysis import (
    assign_data_to_clusters,
    calculate_2d_histogram,
    calculate_density_map,
    find_cutoff_near_value,
    find_first_peak_bounds,
    get_robust_threshold,
    ideal_hist_bins,
    prepare_histogram,
    reassign_fluo_ids_by_majority_vote,
)



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.

__doc__ = "Correction/restoration functions."
__all__ = [
    "drift_correction_time_windows_2d",
    "drift_correction_time_windows_3d",
]

from ._correct import drift_correction_time_windows_2d, drift_correction_time_windows_3d



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import pyqtgraph as pg

__APP_NAME__ = "pyMINFLUX"
__version__ = "0.3.0"

# PyQtGraph settings
pg.setConfigOption("imageAxisOrder", "row-major")  # For best performance

# Documentation
__doc__ = f"""
This is the **development** documentation of the `pyminflux` core command-line API (version {__version__}).

Currently, `pyminflux` is compatible with python 3.10 and 3.11.

**Usage**: see [example notebook](https://github.com/bsse-scf/pyMINFLUX/blob/feature/extended_stats/examples/processing.ipynb).

**Project home**: [https://pyminflux.ethz.ch](https://pyminflux.ethz.ch)
"""

# Do not build the documentation for the following modules that are only used by the user interface.
__pdoc__ = {
    "base": False,
    "main": False,
    "resources": False,
    "settings": False,
    "state": False,
    "threads": False,
    "ui": False,
    "utils": False,
}



# -*- coding: utf-8 -*-

#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

################################################################################
## Form generated from reading UI file 'roi_ranges.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QAbstractButton,
    QApplication,
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)


class Ui_ROIRanges(object):
    def setupUi(self, ROIRanges):
        if not ROIRanges.objectName():
            ROIRanges.setObjectName("ROIRanges")
        ROIRanges.resize(390, 140)
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(ROIRanges.sizePolicy().hasHeightForWidth())
        ROIRanges.setSizePolicy(sizePolicy)
        ROIRanges.setMaximumSize(QSize(16777215, 16777215))
        self.gridLayout = QGridLayout(ROIRanges)
        self.gridLayout.setObjectName("gridLayout")
        self.buttonBox = QDialogButtonBox(ROIRanges)
        self.buttonBox.setObjectName("buttonBox")
        self.buttonBox.setOrientation(Qt.Horizontal)
        self.buttonBox.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)

        self.gridLayout.addWidget(self.buttonBox, 4, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(
            20, 5, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout.addItem(self.verticalSpacer, 3, 0, 1, 1)

        self.efo_layout = QHBoxLayout()
        self.efo_layout.setObjectName("efo_layout")
        self.lbEFO = QLabel(ROIRanges)
        self.lbEFO.setObjectName("lbEFO")

        self.efo_layout.addWidget(self.lbEFO)

        self.efo_spacer = QSpacerItem(40, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.efo_layout.addItem(self.efo_spacer)

        self.lbEFOMin = QLabel(ROIRanges)
        self.lbEFOMin.setObjectName("lbEFOMin")

        self.efo_layout.addWidget(self.lbEFOMin)

        self.leEFOMin = QLineEdit(ROIRanges)
        self.leEFOMin.setObjectName("leEFOMin")

        self.efo_layout.addWidget(self.leEFOMin)

        self.lbEFOMax = QLabel(ROIRanges)
        self.lbEFOMax.setObjectName("lbEFOMax")

        self.efo_layout.addWidget(self.lbEFOMax)

        self.leEFOMax = QLineEdit(ROIRanges)
        self.leEFOMax.setObjectName("leEFOMax")

        self.efo_layout.addWidget(self.leEFOMax)

        self.gridLayout.addLayout(self.efo_layout, 2, 0, 1, 1)

        self.cfr_layout = QHBoxLayout()
        self.cfr_layout.setObjectName("cfr_layout")
        self.lbCFR = QLabel(ROIRanges)
        self.lbCFR.setObjectName("lbCFR")

        self.cfr_layout.addWidget(self.lbCFR)

        self.cfr_spacer = QSpacerItem(40, 0, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.cfr_layout.addItem(self.cfr_spacer)

        self.lbMinCFR = QLabel(ROIRanges)
        self.lbMinCFR.setObjectName("lbMinCFR")

        self.cfr_layout.addWidget(self.lbMinCFR)

        self.leCFRMin = QLineEdit(ROIRanges)
        self.leCFRMin.setObjectName("leCFRMin")

        self.cfr_layout.addWidget(self.leCFRMin)

        self.lbCFRMax = QLabel(ROIRanges)
        self.lbCFRMax.setObjectName("lbCFRMax")

        self.cfr_layout.addWidget(self.lbCFRMax)

        self.leCFRMax = QLineEdit(ROIRanges)
        self.leCFRMax.setObjectName("leCFRMax")

        self.cfr_layout.addWidget(self.leCFRMax)

        self.gridLayout.addLayout(self.cfr_layout, 1, 0, 1, 1)

        self.verticalSpacer_2 = QSpacerItem(
            20, 5, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout.addItem(self.verticalSpacer_2, 0, 0, 1, 1)

        self.retranslateUi(ROIRanges)
        self.buttonBox.accepted.connect(ROIRanges.accept)
        self.buttonBox.rejected.connect(ROIRanges.reject)

        QMetaObject.connectSlotsByName(ROIRanges)

    # setupUi

    def retranslateUi(self, ROIRanges):
        self.lbEFO.setText(QCoreApplication.translate("ROIRanges", "EFO", None))
        self.lbEFOMin.setText(QCoreApplication.translate("ROIRanges", "min", None))
        self.lbEFOMax.setText(QCoreApplication.translate("ROIRanges", "max", None))
        self.lbCFR.setText(QCoreApplication.translate("ROIRanges", "CFR", None))
        self.lbMinCFR.setText(QCoreApplication.translate("ROIRanges", "min", None))
        self.lbCFRMax.setText(QCoreApplication.translate("ROIRanges", "max", None))
        pass

    # retranslateUi



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'color_unmixer.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)


class Ui_ColorUnmixer(object):
    def setupUi(self, ColorUnmixer):
        if not ColorUnmixer.objectName():
            ColorUnmixer.setObjectName("ColorUnmixer")
        ColorUnmixer.resize(800, 600)
        self.verticalLayout_2 = QVBoxLayout(ColorUnmixer)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.main_layout = QVBoxLayout()
        self.main_layout.setObjectName("main_layout")
        self.commands_layout = QHBoxLayout()
        self.commands_layout.setObjectName("commands_layout")
        self.twUnmixingOptions = QTabWidget(ColorUnmixer)
        self.twUnmixingOptions.setObjectName("twUnmixingOptions")
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.twUnmixingOptions.sizePolicy().hasHeightForWidth()
        )
        self.twUnmixingOptions.setSizePolicy(sizePolicy)
        self.tab = QWidget()
        self.tab.setObjectName("tab")
        self.horizontalLayout = QHBoxLayout(self.tab)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.horizontalSpacer = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.lbNumFluorophores = QLabel(self.tab)
        self.lbNumFluorophores.setObjectName("lbNumFluorophores")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(
            self.lbNumFluorophores.sizePolicy().hasHeightForWidth()
        )
        self.lbNumFluorophores.setSizePolicy(sizePolicy1)
        self.lbNumFluorophores.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.lbNumFluorophores)

        self.cbNumFluorophores = QComboBox(self.tab)
        self.cbNumFluorophores.addItem("")
        self.cbNumFluorophores.addItem("")
        self.cbNumFluorophores.setObjectName("cbNumFluorophores")
        sizePolicy2 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.cbNumFluorophores.sizePolicy().hasHeightForWidth()
        )
        self.cbNumFluorophores.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.cbNumFluorophores)

        self.pbDetect = QPushButton(self.tab)
        self.pbDetect.setObjectName("pbDetect")
        sizePolicy2.setHeightForWidth(self.pbDetect.sizePolicy().hasHeightForWidth())
        self.pbDetect.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.pbDetect)

        self.pbAssign = QPushButton(self.tab)
        self.pbAssign.setObjectName("pbAssign")
        sizePolicy2.setHeightForWidth(self.pbAssign.sizePolicy().hasHeightForWidth())
        self.pbAssign.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.pbAssign)

        self.horizontalSpacer_2 = QSpacerItem(
            61, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout.addItem(self.horizontalSpacer_2)

        self.twUnmixingOptions.addTab(self.tab, "")
        self.tab_2 = QWidget()
        self.tab_2.setObjectName("tab_2")
        self.horizontalLayout_3 = QHBoxLayout(self.tab_2)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalSpacer_4 = QSpacerItem(
            40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_2.addItem(self.horizontalSpacer_4)

        self.lbManualThreshold = QLabel(self.tab_2)
        self.lbManualThreshold.setObjectName("lbManualThreshold")

        self.horizontalLayout_2.addWidget(self.lbManualThreshold)

        self.leManualThreshold = QLineEdit(self.tab_2)
        self.leManualThreshold.setObjectName("leManualThreshold")
        sizePolicy2.setHeightForWidth(
            self.leManualThreshold.sizePolicy().hasHeightForWidth()
        )
        self.leManualThreshold.setSizePolicy(sizePolicy2)

        self.horizontalLayout_2.addWidget(self.leManualThreshold)

        self.pbPreview = QPushButton(self.tab_2)
        self.pbPreview.setObjectName("pbPreview")
        sizePolicy3 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.pbPreview.sizePolicy().hasHeightForWidth())
        self.pbPreview.setSizePolicy(sizePolicy3)

        self.horizontalLayout_2.addWidget(self.pbPreview)

        self.pbManualAssign = QPushButton(self.tab_2)
        self.pbManualAssign.setObjectName("pbManualAssign")

        self.horizontalLayout_2.addWidget(self.pbManualAssign)

        self.horizontalSpacer_3 = QSpacerItem(
            61, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_2.addItem(self.horizontalSpacer_3)

        self.horizontalLayout_3.addLayout(self.horizontalLayout_2)

        self.twUnmixingOptions.addTab(self.tab_2, "")
        self.tab_3 = QWidget()
        self.tab_3.setObjectName("tab_3")
        self.horizontalLayout_4 = QHBoxLayout(self.tab_3)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.horizontalSpacer_5 = QSpacerItem(
            241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_4.addItem(self.horizontalSpacer_5)

        self.lbBinSize = QLabel(self.tab_3)
        self.lbBinSize.setObjectName("lbBinSize")
        sizePolicy1.setHeightForWidth(self.lbBinSize.sizePolicy().hasHeightForWidth())
        self.lbBinSize.setSizePolicy(sizePolicy1)

        self.horizontalLayout_4.addWidget(self.lbBinSize)

        self.leBinSize = QLineEdit(self.tab_3)
        self.leBinSize.setObjectName("leBinSize")
        sizePolicy2.setHeightForWidth(self.leBinSize.sizePolicy().hasHeightForWidth())
        self.leBinSize.setSizePolicy(sizePolicy2)

        self.horizontalLayout_4.addWidget(self.leBinSize)

        self.horizontalSpacer_6 = QSpacerItem(
            240, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.horizontalLayout_4.addItem(self.horizontalSpacer_6)

        self.twUnmixingOptions.addTab(self.tab_3, "")

        self.commands_layout.addWidget(self.twUnmixingOptions)

        self.main_layout.addLayout(self.commands_layout)

        self.verticalLayout_2.addLayout(self.main_layout)

        self.retranslateUi(ColorUnmixer)

        self.twUnmixingOptions.setCurrentIndex(0)

        QMetaObject.connectSlotsByName(ColorUnmixer)

    # setupUi

    def retranslateUi(self, ColorUnmixer):
        ColorUnmixer.setWindowTitle(
            QCoreApplication.translate("ColorUnmixer", "Unmixer", None)
        )
        self.lbNumFluorophores.setText(
            QCoreApplication.translate("ColorUnmixer", "Number of fluorophores", None)
        )
        self.cbNumFluorophores.setItemText(
            0, QCoreApplication.translate("ColorUnmixer", "1", None)
        )
        self.cbNumFluorophores.setItemText(
            1, QCoreApplication.translate("ColorUnmixer", "2", None)
        )

        self.pbDetect.setText(
            QCoreApplication.translate("ColorUnmixer", "Detect", None)
        )
        self.pbAssign.setText(
            QCoreApplication.translate("ColorUnmixer", "Assign", None)
        )
        self.twUnmixingOptions.setTabText(
            self.twUnmixingOptions.indexOf(self.tab),
            QCoreApplication.translate("ColorUnmixer", "Automatic", None),
        )
        self.lbManualThreshold.setText(
            QCoreApplication.translate("ColorUnmixer", "DCR threshold", None)
        )
        self.pbPreview.setText(
            QCoreApplication.translate("ColorUnmixer", "Preview", None)
        )
        self.pbManualAssign.setText(
            QCoreApplication.translate("ColorUnmixer", "Assign", None)
        )
        self.twUnmixingOptions.setTabText(
            self.twUnmixingOptions.indexOf(self.tab_2),
            QCoreApplication.translate("ColorUnmixer", "Manual", None),
        )
        self.lbBinSize.setText(
            QCoreApplication.translate(
                "ColorUnmixer", "Bin size (set to 0 for auto)", None
            )
        )
        self.twUnmixingOptions.setTabText(
            self.twUnmixingOptions.indexOf(self.tab_3),
            QCoreApplication.translate("ColorUnmixer", "Histogram settings", None),
        )

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

from PySide6.QtCore import QSignalBlocker, Qt, Signal, Slot
from PySide6.QtWidgets import QWidget

from ..reader import MinFluxReader
from ..state import ColorCode, State
from .ui_plotter_toolbar import Ui_PlotterToolbar


class PlotterToolbar(QWidget, Ui_PlotterToolbar):
    plot_requested_parameters = Signal(None, name="plot_requested_parameters")
    color_code_locs_changed = Signal(int, name="color_code_locs_changed")
    plot_average_positions_state_changed = Signal(
        None, name="plot_average_positions_state_changed"
    )

    def __init__(self):
        """Constructor."""

        # Call the base class
        super().__init__()

        # Initialize the dialog
        self.ui = Ui_PlotterToolbar()
        self.ui.setupUi(self)

        # Initialize state
        self.state = State()

        # The main plotter does not focus on temporal measurement
        self.plotting_parameters = MinFluxReader.processed_properties()

        # Add the values to the plot properties combo boxes (without time)
        self.ui.cbFirstParam.addItems(self.plotting_parameters)
        self.ui.cbFirstParam.setCurrentIndex(self.plotting_parameters.index("x"))
        self.ui.cbSecondParam.addItems(self.plotting_parameters)
        self.ui.cbSecondParam.setCurrentIndex(self.plotting_parameters.index("y"))

        self.ui.cbFirstParam.currentIndexChanged.connect(self.persist_first_param)
        self.ui.cbFirstParam.currentIndexChanged.connect(self.toggle_average_state)
        self.ui.cbSecondParam.currentIndexChanged.connect(self.persist_second_param)
        self.ui.cbSecondParam.currentIndexChanged.connect(self.toggle_average_state)

        # Plot
        self.ui.pbPlot.clicked.connect(self.emit_plot_requested)

        # Color-code combo box
        self.ui.cbColorCodeSelector.setCurrentIndex(0)
        self.ui.cbColorCodeSelector.currentIndexChanged.connect(
            self.persist_color_code_and_broadcast
        )

        # Set the state of the average checkbox
        self.ui.cbPlotAveragePos.setChecked(self.state.plot_average_localisations)
        self.ui.cbPlotAveragePos.stateChanged.connect(
            self.persist_plot_average_localisations_and_broadcast
        )

    @Slot(int, name="persist_plot_average_localisations_and_broadcast")
    def persist_plot_average_localisations_and_broadcast(self, value):
        """Persist the selection for plotting average positions."""
        if value == Qt.CheckState.Checked.value:
            self.state.plot_average_localisations = True
        else:
            self.state.plot_average_localisations = False
        self.plot_average_positions_state_changed.emit()

    @Slot(int, name="toggle_average_state")
    def toggle_average_state(self, index):
        """Persist the selection for the second parameter."""

        # Enable the Average checkbox only for x, y plots
        if self.state.x_param in ["x", "y"] and self.state.y_param in ["x", "y"]:
            self.ui.cbPlotAveragePos.setEnabled(True)
        else:
            self.ui.cbPlotAveragePos.setEnabled(False)

    @Slot(int, name="persist_color_code_and_broadcast")
    def persist_color_code_and_broadcast(self, index):
        """Persist the selection of the color code and broadcast a change."""
        self.state.color_code = ColorCode(index)

        # Broadcast the change
        self.color_code_locs_changed.emit(self.state.color_code.value)

    @Slot(int, name="persist_first_param")
    def persist_first_param(self, index):
        """Persist the selection for the first parameter."""

        # Persist the selection
        self.state.x_param = self.plotting_parameters[index]

    @Slot(int, name="persist_second_param")
    def persist_second_param(self, index):
        """Persist the selection for the second parameter."""

        # Persist the selection
        self.state.y_param = self.plotting_parameters[index]

    @Slot(None, name="emit_plot_requested")
    def emit_plot_requested(self):
        """Plot the requested parameters."""

        # Emit signal
        self.plot_requested_parameters.emit()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#
import os
import warnings

import numpy as np
import pyqtgraph as pg
from PySide6.QtCore import (
    QMutex,
    QMutexLocker,
    QObject,
    QPoint,
    QRunnable,
    QThread,
    QThreadPool,
    Signal,
    Slot,
)
from PySide6.QtGui import QAction, QDoubleValidator, QIntValidator, Qt
from PySide6.QtWidgets import QDialog, QMenu

from ..fourier import estimate_resolution_by_frc
from ..processor import MinFluxProcessor
from ..state import State
from .helpers import export_plot_interactive
from .ui_frc_tool import Ui_FRCTool


class CancelFlag:
    """Encapulates a Mutex to communicate state change to the Workers."""

    def __init__(self):
        self._cancelled = False
        self._mutex = QMutex()

    def is_cancelled(self):
        with QMutexLocker(self._mutex):
            return self._cancelled

    def cancel(self):
        with QMutexLocker(self._mutex):
            self._cancelled = True


class WorkerSignals(QObject):
    """Encapsulates Signals to be used by the Workers."""

    # Signal progress, completion, and request for stop
    progress = Signal(int)
    result = Signal(int, np.ndarray)


class Worker(QRunnable):
    """Worker to run the FRC analysis in parallel over all bins."""

    def __init__(
        self, index, processor, start_time, end_time, sxy, rx, ry, n_reps, cancel_flag
    ):
        """Constructor."""

        super().__init__()
        self.index = index
        self.processor = processor
        self.start_time = start_time
        self.end_time = end_time
        self.sxy = sxy
        self.rx = rx
        self.ry = ry
        self.n_reps = n_reps
        self.signals = WorkerSignals()
        self.cancel_flag = cancel_flag

    def run(self):
        # Check that the processing has not been interrupted
        if self.cancel_flag.is_cancelled():
            # Signal completion
            self.signals.result.emit(self.index, None)
            self.signals.progress.emit(1)
            return

        # Extract the data
        df = self.processor.select_by_1d_range(
            "tim", x_range=(self.start_time, self.end_time)
        )
        df_gr = df.groupby("tid")
        mx = df_gr["x"].mean()
        my = df_gr["y"].mean()

        # Estimate the resolution by FRC
        _, _, _, resolutions, _ = estimate_resolution_by_frc(
            x=mx,
            y=my,
            sx=self.sxy,
            sy=self.sxy,
            rx=self.rx,
            ry=self.ry,
            num_reps=self.n_reps,
            seed=None,
            return_all=True,
        )

        # Convert to nm
        resolutions = 1e9 * np.array(resolutions)

        # Signal completion
        self.signals.result.emit(self.index, resolutions)
        self.signals.progress.emit(1)


class ProcessorThread(QThread):
    # Signal progress and completion
    processing_started = Signal()
    processing_progress_updated = Signal(int)
    processing_interrupted = Signal()
    processing_finished = Signal()

    def __init__(self, processor, t_start, t_steps, sxy, rx, ry, n_reps):
        """Constructor."""
        super().__init__()

        # Set arguments
        self.processor = processor
        self.t_start = t_start
        self.t_steps = t_steps
        self.sxy = sxy
        self.rx = rx
        self.ry = ry
        self.n_reps = n_reps

        # Initialize a thread pool
        self.thread_pool = QThreadPool.globalInstance()
        self.thread_pool.setMaxThreadCount(os.cpu_count())

        # Keep track of the progress of the various threads
        self.progress = 0

        # Initialize output
        self.all_resolutions = None

        # Initialize CancelFlag to communicate to Worker in a thread-safe manner
        self.cancel_flag = CancelFlag()

    def run(self):
        """Run the computation."""

        # Signal start
        self.processing_started.emit()

        # Allocate space for the results
        self.all_resolutions = np.empty(
            (len(self.t_steps), self.n_reps), dtype=np.float32
        )
        self.all_resolutions.fill(np.nan)

        # Reset the progress counter
        self.progress = 0

        # Create the workers and add them to the thread pool
        for i, t in enumerate(self.t_steps):
            if self.cancel_flag.is_cancelled():
                # This is unlikely: threads are added very fast to the QThreadPool's queue
                self.processing_interrupted.emit()
                break

            # Create the processor_thread
            worker = Worker(
                index=i,
                processor=self.processor,
                start_time=self.t_start,
                end_time=t,
                sxy=self.sxy,
                rx=self.rx,
                ry=self.ry,
                n_reps=self.n_reps,
                cancel_flag=self.cancel_flag,
            )

            # Attach all signals
            worker.signals.result.connect(self.store_result)
            worker.signals.progress.connect(self.broadcast_update_progress)

            # Add the thread to the pool and start it
            self.thread_pool.start(
                worker
            )  # Add the processor_thread to the thread pool

        # Wait for all threads to finish
        self.thread_pool.waitForDone()

        # Signal completion
        self.processing_finished.emit()

    def stop(self):
        # Inform the queued Workers to skip computation
        self.cancel_flag.cancel()

        # Clear the ThreadPool queue
        self.thread_pool.clear()

    @Slot(int, np.ndarray, name="store_result")
    def store_result(self, index, result):
        if result is not None:
            self.all_resolutions[index, :] = result

    def broadcast_update_progress(self, value):
        # Signal progress
        self.progress += value
        self.processing_progress_updated.emit(
            int(np.round(100 * (self.progress + 1) / len(self.t_steps)))
        )


class FRCTool(QDialog, Ui_FRCTool):
    def __init__(self, processor: MinFluxProcessor, parent=None):
        """Constructor."""

        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_FRCTool()
        self.ui.setupUi(self)

        # Store the reference to the reader
        self.processor = processor

        # Keep reference to the plot
        self.frc_plot = None

        # Keep a reference to the singleton State class
        self.state = State()

        # Create widgets
        self.create_widgets()

        # Keep a reference to the processor thread
        self.processor_thread = None

        # Hide the progress bar
        self.ui.progress_bar.setValue(0)
        self.ui.progress_bar.setVisible(False)
        self.ui.pbInterrupt.setVisible(False)

        # Keep track of whether there is a plot to export
        self.plot_ready_to_export = False

        #
        # Set defaults
        #

        # Lateral (xy) resolution
        self.ui.leLateralResolution.setText(str(self.state.frc_lateral_resolution))
        self.ui.leNumRepeats.setValidator(QDoubleValidator(bottom=0.0, decimals=2))

        # Temporal (s) resolution
        self.ui.leTemporalResolution.setText(str(self.state.frc_temporal_resolution))
        self.ui.leTemporalResolution.setValidator(
            QDoubleValidator(bottom=0.0, decimals=1)
        )

        # Endpoint estimation
        self.ui.cbEndpoint.setChecked(self.state.frc_endpoint_only)
        self.ui.lbTemporalResolution.setEnabled(True)
        self.ui.leTemporalResolution.setEnabled(True)
        if self.state.frc_endpoint_only:
            self.ui.lbTemporalResolution.setEnabled(False)
            self.ui.leTemporalResolution.setEnabled(False)

        # Number of repeats
        self.ui.leNumRepeats.setText(str(self.state.frc_num_repeats))
        self.ui.leNumRepeats.setValidator(QIntValidator(bottom=0))

        # Set signal-slot connections
        self.setup_conn()

    def setup_conn(self):
        """Set up signal-slot connections."""

        self.ui.pbRunFRCAnalysis.clicked.connect(self.run_frc_analysis)
        self.ui.pbInterrupt.clicked.connect(self.stop_frc_analysis)
        self.ui.leLateralResolution.textChanged.connect(
            self.persist_frc_lateral_resolution
        )
        self.ui.leTemporalResolution.textChanged.connect(
            self.persist_frc_temporal_resolution
        )
        self.ui.cbEndpoint.stateChanged.connect(self.persist_frc_endpoint_only)

        self.ui.leNumRepeats.textChanged.connect(self.persist_frc_num_repeats)

    def create_widgets(self):
        """Create widgets."""

        # Parameters Layout
        self.frc_plot = pg.PlotWidget(parent=self, background="w", title="")
        self.frc_plot.setMouseEnabled(x=True, y=True)
        self.frc_plot.setMenuEnabled(False)
        self.frc_plot.hideAxis("bottom")
        self.frc_plot.hideAxis("left")
        self.frc_plot.scene().sigMouseClicked.connect(self.raise_context_menu)
        self.ui.hlPlot.addWidget(self.frc_plot)
        self.frc_plot.show()

    def keyPressEvent(self, ev):
        """Key press event."""

        # Do not capture key events for now
        ev.ignore()

    def set_processor(self, processor: MinFluxProcessor):
        """Reset the internal state."""
        self.processor = processor

    @Slot(int, name="persist_frc_endpoint_only")
    def persist_frc_endpoint_only(self, value):
        """Persist the selection for plotting average positions."""
        if value == Qt.CheckState.Checked.value:
            self.state.frc_endpoint_only = True
            self.ui.lbTemporalResolution.setEnabled(False)
            self.ui.leTemporalResolution.setEnabled(False)
        else:
            self.state.frc_endpoint_only = False
            self.ui.lbTemporalResolution.setEnabled(True)
            self.ui.leTemporalResolution.setEnabled(True)

    @Slot(name="run_frc_analysis")
    def run_frc_analysis(self):
        """Run FRC analysis to estimate signal resolution."""

        # Mark that there is no plot ready to export
        self.plot_ready_to_export = False

        # Get the common parameters
        sxy = self.state.frc_lateral_resolution
        n_reps = self.state.frc_num_repeats

        # Set the spatial ranges
        rx = (
            self.processor.filtered_dataframe["x"].min(),
            self.processor.filtered_dataframe["x"].max(),
        )
        ry = (
            self.processor.filtered_dataframe["y"].min(),
            self.processor.filtered_dataframe["y"].max(),
        )

        # Prepare the arguments for the Worker
        t_start = self.processor.filtered_dataframe["tim"].min()
        t_end = self.processor.filtered_dataframe["tim"].max()

        # Run temporal analysis or endpoint estimation?
        if self.state.frc_endpoint_only:
            # We only consider one step with the whole data
            t_steps = [t_end + 1.0]

        else:
            # Create the list of time steps to consider
            t_step = self.state.frc_temporal_resolution
            n_steps = int(np.round((t_end - t_start) / t_step))
            t_steps = np.linspace(t_start, t_end, n_steps)
            t_steps = t_steps[1:]

        # Create the new processor_thread
        if self.processor_thread is not None:
            del self.processor_thread
        self.processor_thread = ProcessorThread(
            self.processor, t_start, t_steps, sxy, rx, ry, n_reps
        )
        self.processor_thread.processing_started.connect(
            self.update_ui_on_process_start
        )
        self.processor_thread.processing_progress_updated.connect(
            self.update_progress_bar
        )
        self.processor_thread.processing_interrupted.connect(
            self.update_ui_on_process_end
        )
        self.processor_thread.processing_interrupted.connect(
            self.collect_results_and_plot
        )
        self.processor_thread.processing_finished.connect(self.update_ui_on_process_end)
        self.processor_thread.processing_finished.connect(self.collect_results_and_plot)

        # Now process in the local processor_thread
        self.processor_thread.start()

    @Slot(name="stop_frc_analysis")
    def stop_frc_analysis(self):
        if self.processor_thread is None:
            return
        self.processor_thread.stop()

    @Slot(str, name="persist_frc_lateral_resolution")
    def persist_frc_lateral_resolution(self, text):
        try:
            frc_lateral_resolution = float(text)
        except ValueError as _:
            return
        self.state.frc_lateral_resolution = frc_lateral_resolution

    @Slot(str, name="persist_frc_temporal_resolution")
    def persist_frc_temporal_resolution(self, text):
        try:
            frc_temporal_resolution = float(text)
        except ValueError as _:
            return
        self.state.frc_temporal_resolution = frc_temporal_resolution

    @Slot(str, name="persist_frc_num_repeats")
    def persist_frc_num_repeats(self, text):
        try:
            frc_num_repeats = int(text)
        except ValueError as _:
            return
        self.state.frc_num_repeats = frc_num_repeats

    @Slot(int, name="update_progress_bar")
    def update_progress_bar(self, value):
        self.ui.progress_bar.setValue(value)

    def plot(self, time_steps, resolutions):
        """Plot histograms."""

        # Clear the plot
        self.clear_plot()

        # Show x and y grids
        self.frc_plot.showGrid(x=True, y=True)

        # Set properties of the label for y-axis
        self.frc_plot.setLabel("left", "Resolution [nm]")

        # Set properties of the label for x-axis
        self.frc_plot.setLabel("bottom", "Acquisition time", units="s")

        if self.state.frc_endpoint_only:
            # Time steps to use
            time_steps = [time_steps[0] - 1.0]

        # Calculate the mean and standard error of the replicates
        with warnings.catch_warnings():
            # Silence the warning on np.nanmean() for the case that all
            # columns in one or more rows are NaN because the user
            # interrupted the processing.
            warnings.simplefilter("ignore")
            mean_res = np.nanmean(resolutions, axis=1)
            std_res = np.nanstd(resolutions, axis=1, ddof=1)
            se_res = std_res / np.sqrt(resolutions.shape[1])

        # Plot resolution vs. time step
        if len(time_steps) > 1:
            self.frc_plot.plot(time_steps, mean_res, pen=pg.mkPen("r", width=5))
            # Create custom error bars
            for ti, ri, ei in zip(time_steps, mean_res, se_res):
                if np.isnan(ri):
                    continue
                v_line = pg.PlotCurveItem(
                    [ti, ti], [ri - ei, ri + ei], pen=pg.mkPen(color="k", width=3)
                )
                self.frc_plot.addItem(v_line)

            # Set plot x range
            x_range = [time_steps[0], time_steps[-1]]

            # Extract the endpoint resolution (or the last
            # valid if the processing was interrupted)
            res = mean_res[np.isfinite(mean_res)][-1]
            se = se_res[np.isfinite(mean_res)][-1]
        else:
            r = np.nanmean(resolutions, axis=1)[0]
            self.frc_plot.plot(
                [time_steps[0] - 0.1, time_steps[0] + 0.1],
                [r, r],
                pen=pg.mkPen("r", width=5),
            )
            # Create custom error bar
            line = pg.PlotCurveItem(
                [time_steps[0], time_steps[0]],
                [mean_res[0] - se_res[0], mean_res[0] + se_res[0]],
                pen=pg.mkPen(color="k", width=3),
            )
            self.frc_plot.addItem(line)

            # Set plot x range
            x_range = [time_steps[0] - 0.5, time_steps[0] + 0.5]

            # Endpoint resolution
            res = mean_res[0]
            se = se_res[0]

        # Calculate an appropriate y-axis range
        mx = np.nanmax(mean_res + se_res)
        mn = np.nanmin(mean_res - se_res)
        margin = 0.1 * (mx - mn)

        # Set axis ranges
        self.frc_plot.setXRange(x_range[0], x_range[1])
        self.frc_plot.setYRange(mn - margin, mx + margin)

        # Display the endpoint resolution in the plot title
        if np.any(np.isnan(mean_res)):
            self.frc_plot.setTitle(
                f"Endpoint resolution (interrupted) = {res:.1f} ± {se:.1f}nm",
                size="16px",
            )
        else:
            self.frc_plot.setTitle(
                f"Endpoint resolution = {res:.1f} ± {se:.1f}nm", size="16px"
            )

    def clear_plot(self):
        """Clear the plot."""

        for item in self.frc_plot.allChildItems():
            self.frc_plot.removeItem(item)

    def update_ui_on_process_start(self):
        """Disable elements and inform that a processing is ongoing."""

        # Disable all parameter widgets
        self.ui.lbLateralResolution.setEnabled(False)
        self.ui.leLateralResolution.setEnabled(False)
        self.ui.cbEndpoint.setEnabled(False)
        self.ui.lbTemporalResolution.setEnabled(False)
        self.ui.leTemporalResolution.setEnabled(False)
        self.ui.lbNumRepeats.setEnabled(False)
        self.ui.leNumRepeats.setEnabled(False)
        self.ui.pbRunFRCAnalysis.setEnabled(False)

        # If needed, we make the progress bar visible
        if not self.state.frc_endpoint_only:
            self.ui.progress_bar.setValue(0)
            self.ui.progress_bar.setVisible(True)
            self.ui.pbInterrupt.setVisible(True)

    def update_ui_on_process_end(self):
        """Disable elements and inform that a processing is ongoing."""

        # Enable all parameter widgets
        self.ui.lbLateralResolution.setEnabled(True)
        self.ui.leLateralResolution.setEnabled(True)
        self.ui.cbEndpoint.setEnabled(True)
        self.ui.lbTemporalResolution.setEnabled(True)
        self.ui.leTemporalResolution.setEnabled(True)
        self.ui.lbNumRepeats.setEnabled(True)
        self.ui.leNumRepeats.setEnabled(True)
        self.ui.pbRunFRCAnalysis.setEnabled(True)

        # However, we make the progress bar invisible
        self.ui.progress_bar.setValue(0)
        self.ui.progress_bar.setVisible(False)
        self.ui.pbInterrupt.setVisible(False)

    def collect_results_and_plot(self):
        """Collect results from the local worked and plot them."""

        if self.processor_thread is None:
            return

        # Collect the results
        t_steps = self.processor_thread.t_steps
        all_resolutions = self.processor_thread.all_resolutions

        # Update plot
        self.plot(t_steps, all_resolutions)

        # Mark that there is a plot ready  to export
        self.plot_ready_to_export = True

    def raise_context_menu(self, ev):
        """Create a context menu on the plot."""
        if ev.button() == Qt.MouseButton.RightButton:
            if not self.plot_ready_to_export:
                return

            menu = QMenu()
            export_action = QAction("Export plot")
            export_action.triggered.connect(
                lambda checked: export_plot_interactive(ev.currentItem)
            )
            menu.addAction(export_action)
            pos = ev.screenPos()
            menu.exec(QPoint(int(pos.x()), int(pos.y())))
            ev.accept()
        else:
            ev.ignore()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pyqtgraph as pg
from pyqtgraph import BarGraphItem, PlotWidget
from PySide6 import QtCore, QtWidgets
from PySide6.QtCore import QPoint, Slot
from PySide6.QtGui import QAction, Qt
from PySide6.QtWidgets import QDialog, QMenu

from pyminflux.analysis import prepare_histogram
from pyminflux.processor import MinFluxProcessor
from pyminflux.state import State
from pyminflux.ui.helpers import export_plot_interactive
from pyminflux.ui.ui_trace_length_viewer import Ui_TraceLengthViewer


class TraceLengthViewer(QDialog, Ui_TraceLengthViewer):
    """
    A QDialog to display a bar plot of trace lengths.
    """

    def __init__(self, processor: MinFluxProcessor, parent):
        # Call the base class
        super().__init__(parent=parent)

        # Initialize the dialog
        self.ui = Ui_TraceLengthViewer()
        self.ui.setupUi(self)

        # Initialize state
        self.state = State()

        # Constants
        self.brush = pg.mkBrush(0, 0, 0, 255)
        self.pen = pg.mkPen(None)

        # Keep a reference to the Processor
        self.processor = processor

        # Create the plot elements
        self.plot_widget = PlotWidget(
            parent=self, background="w", title="Trace Length Distribution"
        )

        # Plot the dcr histogram
        self.plot_trace_lengths()

        # Add them to the UI
        self.ui.main_layout.addWidget(self.plot_widget)

    @Slot(None, name="plot_dcr_histogram")
    def plot_trace_lengths(self):
        """Plot the trace length."""

        # Do we have something to plot?
        if self.processor is None or self.processor.full_dataframe is None:
            return

        # Prepare the histogram to plot
        all_counts, _, all_trace_lengths, _ = prepare_histogram(
            self.processor.filtered_dataframe_stats["n"],
            normalize=False,
            auto_bins=False,
            bin_size=1.0,
        )

        # Remove empty trace lengths
        to_keep = all_counts > 0
        counts = all_counts[to_keep]
        trace_lengths = all_trace_lengths[to_keep]

        # Set x and y ranges
        x_range = (0, trace_lengths.max())
        y_range = (0, counts.max())

        # It one or more charts already exists, remove them
        for item in self.plot_widget.allChildItems():
            self.plot_widget.removeItem(item)

        # Calculate median trace length
        median_trace_length = np.median(self.processor.filtered_dataframe_stats["n"])

        # Create the bar chart
        chart = BarGraphItem(
            x=trace_lengths,
            height=counts,
            width=0.9,
            brush=self.brush,
        )
        self.plot_widget.setMouseEnabled(x=True, y=False)
        self.plot_widget.setXRange(x_range[0], x_range[1])
        self.plot_widget.setYRange(y_range[0], y_range[1])
        self.plot_widget.setLabel("bottom", "Trace length")
        self.plot_widget.setLabel("left", "Trace count")
        self.plot_widget.setMenuEnabled(False)
        self.plot_widget.addItem(chart)
        self.plot_widget.setTitle(
            f"Median trace length = {median_trace_length:.1f} localizations"
        )
        self.plot_widget.scene().sigMouseClicked.connect(
            self.histogram_raise_context_menu
        )

        # Add the mean value as a vertical red line
        line = pg.InfiniteLine(
            pos=median_trace_length,
            movable=False,
            angle=90,
            pen={"color": (200, 50, 50), "width": 3},
            label=f"median={median_trace_length:.1f}",
            labelOpts={
                "position": 0.95,
                "color": (200, 50, 50),
                "fill": (200, 50, 50, 10),
                "movable": True,
            },
        )
        self.plot_widget.addItem(line)

    def histogram_raise_context_menu(self, ev):
        """Create a context menu on the plot."""
        if ev.button() == Qt.MouseButton.RightButton:
            menu = QMenu()
            export_action = QAction("Export plot")
            export_action.triggered.connect(
                lambda checked: export_plot_interactive(ev.currentItem)
            )
            menu.addAction(export_action)
            pos = ev.screenPos()
            menu.exec(QPoint(int(pos.x()), int(pos.y())))
            ev.accept()
        else:
            ev.ignore()

    @Slot(None, name="update")
    def update(self):
        """Update the viewer."""
        if self.processor is None:
            return
        self.plot_trace_lengths()



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Optional

import numpy as np


def render_xy(
    x,
    y,
    sx: float = 1.0,
    sy: float = 1.0,
    rx: Optional[tuple] = None,
    ry: Optional[tuple] = None,
    render_type: Optional[str] = "histogram",
    fwhm: Optional[float] = None,
):
    """Renders the localizations as a 2D image with given resolution.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    x: np.ndarray
        Array of localization x coordinates.

    y: np.ndarray
        Array of localization y coordinates.

    sx: float (Default = 1.0)
        Resolution of the render in the x direction.

    sy: float (Default = 1.0)
        Resolution of the render in the y direction.

    rx: tuple (Optional)
        (min, max) boundaries for the x coordinates. If omitted, it will default to (x.min(), x.max()).

    ry: float (Optional)
        (min, max) boundaries for the y coordinates. If omitted, it will default to (y.min(), y.max()).

    render_type: str (Default = "histogram")
        Type of render to be generated. It must be one of:
            "histogram": simple 2D histogram of localization falling into each bin of size (sx, sy).
            "fixed_gaussian": sub-pixel resolution Gaussian fit. The Gaussian full-width half maximum is required
            (see below).

    fwhm: float (Optional)
        Requested full-width half maximum (FWHM) of the Gaussian kernel. If omitted, it is set to be
        3 * np.sqrt(np.power(sx, 2) + np.power(sy, 2)).

    Returns
    -------

    h: np.ndarray
        Rendered image (as float32 2D NumPy array)
    xi: np.ndarray
        Array of x coordinates for the output x, y grid
    yi: np.ndarray
        Array of x coordinates for the output x, y grid
    m: np.ndarray
        Logical array with the positions that were considered. The False entries were rejected because they were
        outside the rx, ry ranges (with the additional constraint of the edge effect of the Gaussian support for
        the "fixed_gaussian" render type.
    """

    if render_type is None:
        render_type = "histogram"

    render_type = render_type.lower()
    if render_type not in ["histogram", "fixed_gaussian"]:
        raise ValueError("plot_type must be one of 'histogram' or 'fixed_gaussian'.'")

    if render_type == "fixed_gaussian" and fwhm is None:
        sxy = np.sqrt(np.power(sx, 2) + np.power(sy, 2))
        fwhm = 3 * sxy

    # Make sure we are working with NumPy arrays
    x = np.array(x)
    y = np.array(y)

    # Make sure rx and ry are defined
    if rx is None:
        rx = (x.min(), x.max())
    if ry is None:
        ry = (y.min(), y.max())

    # Get dimensions and allocate output array
    Nx = int(np.ceil((rx[1] - rx[0]) / sx))
    Ny = int(np.ceil((ry[1] - ry[0]) / sy))
    h = np.zeros((Ny, Nx), dtype=np.float32)

    # Get position in pixels
    px = (x - rx[0]) / sx
    py = (y - ry[0]) / sy

    # Convert absolute position to image indices
    ix = np.round(px).astype(int)
    iy = np.round(py).astype(int)

    # Remove positions outside the image
    m = (ix >= 0) & (ix < Nx) & (iy >= 0) & (iy < Ny)
    px = px[m]
    py = py[m]
    ix = ix[m]
    iy = iy[m]

    # Plot requested image type
    if render_type == "histogram":
        # Fill in histogram
        for i in range(len(ix)):
            xi = ix[i]
            yi = Ny - iy[i] - 1  # Images have y = 0 at the top
            try:
                h[yi, xi] += 1
            except IndexError as e:
                print(
                    f"Tried to access (y={yi}, x={xi}) in image of size (Ny={Ny}, Nx={Nx})."
                )

    elif render_type == "fixed_gaussian":
        # Gaussian with subpixel accuracy
        wx = fwhm / sx
        wy = fwhm / sy
        L = int(np.ceil(2 * max(wx, wy)))

        # Small grid
        g = np.arange(-L, L + 1)
        yk, xk = np.meshgrid(g, g)

        # Remove close to borders
        m = (ix >= L) & (ix < Nx - L) & (iy >= L) & (iy < Ny - L)
        px = px[m]
        py = py[m]
        ix = ix[m]
        iy = iy[m]

        for i in range(len(ix)):
            xi = ix[i]
            yi = iy[i]
            dx = px[i] - xi
            dy = py[i] - yi
            gx = xi + g
            gy = yi + g

            # Calculate the Gaussian kernel using the requested FWHM.
            k = np.exp(
                -4 * np.log(2) * ((xk - dx) ** 2 / wx**2 + (yk - dy) ** 2 / wy**2)
            )

            # Add it to the image
            my, mx = np.meshgrid(
                gy, gx, indexing="ij"
            )  # We need to create meshgrid to add the k matrix
            my = Ny - my - 1  # Images have y = 0 at the top
            try:
                h[my, mx] = h[my, mx] + k
            except IndexError as e:
                print(
                    f"Tried to access (y={yi}, x={xi}) in image of size (Ny={Ny}, Nx={Nx})."
                )

    else:
        raise ValueError("Unknown type")

    # Define output xy grid
    xi = rx[0] + (np.arange(Nx)) * sx + sx / 2
    yi = ry[0] + (np.arange(Ny)) * sy + sy / 2

    return h, xi, yi, m


def render_xyz(
    x,
    y,
    z,
    sx: float = 1.0,
    sy: float = 1.0,
    sz: float = 1.0,
    rx: Optional[tuple] = None,
    ry: Optional[tuple] = None,
    rz: Optional[tuple] = None,
    render_type: Optional[str] = "histogram",
    fwhm: Optional[float] = None,
):
    """Renders the localizations as a 3D stack with given resolution.

    Reimplemented (with modifications) from:

    * [paper] Ostersehlt, L.M., Jans, D.C., Wittek, A. et al. DNA-PAINT MINFLUX nanoscopy. Nat Methods 19, 1072-1075 (2022). https://doi.org/10.1038/s41592-022-01577-1
    * [code]  https://zenodo.org/record/6563100

    Parameters
    ----------

    x: np.ndarray
        Array of localization x coordinates.

    y: np.ndarray
        Array of localization y coordinates.

    z: np.ndarray
        Array of localization z coordinates.

    sx: float (Default = 1.0)
        Resolution of the render in the x direction.

    sy: float (Default = 1.0)
        Resolution of the render in the y direction.

    sz: float (Default = 1.0)
        Resolution of the render in the z direction.

    rx: tuple (Optional)
        (min, max) boundaries for the x coordinates. If omitted, it will default to (x.min(), x.max()).

    ry: float (Optional)
        (min, max) boundaries for the y coordinates. If omitted, it will default to (y.min(), y.max()).

    rz: float (Optional)
        (min, max) boundaries for the z coordinates. If omitted, it will default to (z.min(), z.max()).

    render_type: str (Default = "histogram")
        Type of render to be generated. It must be one of:
            "histogram": simple 2D histogram of localization falling into each bin of size (sx, sy).
            "fixed_gaussian": sub-pixel resolution Gaussian fit. The Gaussian full-width half maximum is required
            (see below).

    fwhm: float (Optional)
        Requested full-width half maximum (FWHM) of the Gaussian kernel. If omitted, it is set to be
        3 * np.sqrt(np.power(sx, 2) + np.power(sy, 2)).

    Returns
    -------

    h: np.ndarray
        Rendered image (as float32 2D NumPy array)
    xi: np.ndarray
        Array of x coordinates for the output x, y grid
    yi: np.ndarray
        Array of x coordinates for the output x, y grid
    m: np.ndarray
        Logical array with the positions that were considered. The False entries were rejected because they were
        outside the rx, ry ranges (with the additional constraint of the edge effect of the Gaussian support for
        the "fixed_gaussian" render type.
    """

    if render_type is None:
        render_type = "histogram"

    render_type = render_type.lower()
    if render_type not in ["histogram", "fixed_gaussian"]:
        raise ValueError("plot_type must be one of 'histogram' or 'fixed_gaussian'.'")

    if render_type == "fixed_gaussian" and fwhm is None:
        sxyz = np.sqrt(np.power(sx, 2) + np.power(sy, 2) + np.power(sz, 2))
        fwhm = 3 * sxyz

    # Make sure we are working with NumPy arrays
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)

    # Make sure rx and ry are defined
    if rx is None:
        rx = (x.min(), x.max())
    if ry is None:
        ry = (y.min(), y.max())
    if rz is None:
        rz = (z.min(), z.max())

    # Get dimensions and allocate output array
    Nx = int(np.ceil((rx[1] - rx[0]) / sx))
    Ny = int(np.ceil((ry[1] - ry[0]) / sy))
    Nz = int(np.ceil((rz[1] - rz[0]) / sz))
    h = np.zeros((Nz, Ny, Nx), dtype=np.float32)

    # Get position in pixels
    px = (x - rx[0]) / sx
    py = (y - ry[0]) / sy
    pz = (z - rz[0]) / sz

    # Convert absolute position to image indices
    ix = np.round(px).astype(int)
    iy = np.round(py).astype(int)
    iz = np.round(pz).astype(int)

    # Remove positions outside the image
    m = (ix >= 0) & (ix < Nx) & (iy >= 0) & (iy < Ny) & (iz >= 0) & (iz < Nz)
    px = px[m]
    py = py[m]
    pz = pz[m]
    ix = ix[m]
    iy = iy[m]
    iz = iz[m]

    # Plot requested image type
    if render_type == "histogram":
        # Fill in histogram
        for i in range(len(ix)):
            xi = ix[i]
            yi = Ny - iy[i] - 1  # Images have y = 0 at the top
            zi = iz[i]
            try:
                h[zi, yi, xi] += 1
            except IndexError as e:
                print(
                    f"Tried to access (z={zi}, y={yi}, x={xi}) in image of size (Nz={Nz}, Ny={Ny}, Nx={Nx})."
                )

    elif render_type == "fixed_gaussian":
        # Gaussian with subpixel accuracy
        wx = fwhm / sx
        wy = fwhm / sy
        wz = fwhm / sz
        L = int(np.ceil(2 * max(wx, wy, wz)))

        # Small grid
        g = np.arange(-L, L + 1)
        zk, yk, xk = np.meshgrid(g, g, g, indexing="ij")

        # Remove close to borders
        m = (
            (ix >= L)
            & (ix < Nx - L)
            & (iy >= L)
            & (iy < Ny - L)
            & (iz >= L)
            & (iz < Nz - L)
        )
        px = px[m]
        py = py[m]
        pz = pz[m]
        ix = ix[m]
        iy = iy[m]
        iz = iz[m]

        for i in range(len(ix)):
            xi = ix[i]
            yi = iy[i]
            zi = iz[i]
            dx = px[i] - xi
            dy = py[i] - yi
            dz = pz[i] - zi
            gx = xi + g
            gy = yi + g
            gz = zi + g

            # Calculate the Gaussian kernel using the requested FWHM.
            k = np.exp(
                -4
                * np.log(2)
                * (
                    (xk - dx) ** 2 / wx**2
                    + (yk - dy) ** 2 / wy**2
                    + (zk - dz) ** 2 / wz**2
                )
            )

            # Add it to the image
            mz, my, mx = np.meshgrid(
                gz, gy, gx, indexing="ij"
            )  # We need to create meshgrid to add the k matrix
            my = Ny - my - 1  # Images have y = 0 at the top
            try:
                h[mz, my, mx] = h[mz, my, mx] + k
            except IndexError as e:
                print(
                    f"Tried to access (y={yi}, x={xi}) in image of size (Ny={Ny}, Nx={Nx})."
                )

    else:
        raise ValueError("Unknown type")

    # Define output xy grid
    xi = rx[0] + (np.arange(Nx)) * sx + sx / 2
    yi = ry[0] + (np.arange(Ny)) * sy + sy / 2
    zi = rz[0] + (np.arange(Nz)) * sz + sz / 2

    return h, xi, yi, zi, m



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'wizard.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QComboBox,
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QVBoxLayout,
    QWidget,
)


class Ui_WizardDialog(object):
    def setupUi(self, WizardDialog):
        if not WizardDialog.objectName():
            WizardDialog.setObjectName("WizardDialog")
        WizardDialog.resize(357, 939)
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.MinimumExpanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(WizardDialog.sizePolicy().hasHeightForWidth())
        WizardDialog.setSizePolicy(sizePolicy)
        self.gridLayout = QGridLayout(WizardDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.verticalSpacer = QSpacerItem(
            20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding
        )

        self.gridLayout.addItem(self.verticalSpacer, 17, 0, 1, 1)

        self.mainLayout = QVBoxLayout()
        self.mainLayout.setObjectName("mainLayout")
        self.pbLoadData = QPushButton(WizardDialog)
        self.pbLoadData.setObjectName("pbLoadData")
        self.pbLoadData.setMinimumSize(QSize(320, 0))

        self.mainLayout.addWidget(self.pbLoadData)

        self.pbReset = QPushButton(WizardDialog)
        self.pbReset.setObjectName("pbReset")

        self.mainLayout.addWidget(self.pbReset)

        self.verticalSpacer_2 = QSpacerItem(
            20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed
        )

        self.mainLayout.addItem(self.verticalSpacer_2)

        self.lbColors = QLabel(WizardDialog)
        self.lbColors.setObjectName("lbColors")
        font = QFont()
        font.setBold(True)
        self.lbColors.setFont(font)

        self.mainLayout.addWidget(self.lbColors)

        self.color_layout = QHBoxLayout()
        self.color_layout.setObjectName("color_layout")
        self.pbSingleColor = QPushButton(WizardDialog)
        self.pbSingleColor.setObjectName("pbSingleColor")

        self.color_layout.addWidget(self.pbSingleColor)

        self.pbColorUnmixer = QPushButton(WizardDialog)
        self.pbColorUnmixer.setObjectName("pbColorUnmixer")

        self.color_layout.addWidget(self.pbColorUnmixer)

        self.mainLayout.addLayout(self.color_layout)

        self.fluorophores_layout = QHBoxLayout()
        self.fluorophores_layout.setObjectName("fluorophores_layout")
        self.lbActiveColor = QLabel(WizardDialog)
        self.lbActiveColor.setObjectName("lbActiveColor")
        sizePolicy1 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(
            self.lbActiveColor.sizePolicy().hasHeightForWidth()
        )
        self.lbActiveColor.setSizePolicy(sizePolicy1)

        self.fluorophores_layout.addWidget(self.lbActiveColor)

        self.cmActiveColor = QComboBox(WizardDialog)
        self.cmActiveColor.addItem("")
        self.cmActiveColor.setObjectName("cmActiveColor")
        sizePolicy2 = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(
            self.cmActiveColor.sizePolicy().hasHeightForWidth()
        )
        self.cmActiveColor.setSizePolicy(sizePolicy2)

        self.fluorophores_layout.addWidget(self.cmActiveColor)

        self.mainLayout.addLayout(self.fluorophores_layout)

        self.verticalSpacer_4 = QSpacerItem(
            20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed
        )

        self.mainLayout.addItem(self.verticalSpacer_4)

        self.lbFilters = QLabel(WizardDialog)
        self.lbFilters.setObjectName("lbFilters")
        self.lbFilters.setFont(font)

        self.mainLayout.addWidget(self.lbFilters)

        self.pbTimeInspector = QPushButton(WizardDialog)
        self.pbTimeInspector.setObjectName("pbTimeInspector")

        self.mainLayout.addWidget(self.pbTimeInspector)

        self.verticalSpacer_3 = QSpacerItem(
            20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed
        )

        self.mainLayout.addItem(self.verticalSpacer_3)

        self.pbAnalyzer = QPushButton(WizardDialog)
        self.pbAnalyzer.setObjectName("pbAnalyzer")

        self.mainLayout.addWidget(self.pbAnalyzer)

        self.lbEFOFiltering = QLabel(WizardDialog)
        self.lbEFOFiltering.setObjectName("lbEFOFiltering")
        self.lbEFOFiltering.setFont(font)

        self.mainLayout.addWidget(self.lbEFOFiltering)

        self.efo_lower_bound_layout = QHBoxLayout()
        self.efo_lower_bound_layout.setObjectName("efo_lower_bound_layout")
        self.lbEFOLowerBound = QLabel(WizardDialog)
        self.lbEFOLowerBound.setObjectName("lbEFOLowerBound")

        self.efo_lower_bound_layout.addWidget(self.lbEFOLowerBound)

        self.leEFOLowerBound = QLineEdit(WizardDialog)
        self.leEFOLowerBound.setObjectName("leEFOLowerBound")
        sizePolicy3 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(
            self.leEFOLowerBound.sizePolicy().hasHeightForWidth()
        )
        self.leEFOLowerBound.setSizePolicy(sizePolicy3)

        self.efo_lower_bound_layout.addWidget(self.leEFOLowerBound)

        self.mainLayout.addLayout(self.efo_lower_bound_layout)

        self.efo_upper_bound_layout = QHBoxLayout()
        self.efo_upper_bound_layout.setObjectName("efo_upper_bound_layout")
        self.lbEFOUpperBound = QLabel(WizardDialog)
        self.lbEFOUpperBound.setObjectName("lbEFOUpperBound")

        self.efo_upper_bound_layout.addWidget(self.lbEFOUpperBound)

        self.leEFOUpperBound = QLineEdit(WizardDialog)
        self.leEFOUpperBound.setObjectName("leEFOUpperBound")
        sizePolicy3.setHeightForWidth(
            self.leEFOUpperBound.sizePolicy().hasHeightForWidth()
        )
        self.leEFOUpperBound.setSizePolicy(sizePolicy3)

        self.efo_upper_bound_layout.addWidget(self.leEFOUpperBound)

        self.mainLayout.addLayout(self.efo_upper_bound_layout)

        self.pbEFOFilter = QPushButton(WizardDialog)
        self.pbEFOFilter.setObjectName("pbEFOFilter")

        self.mainLayout.addWidget(self.pbEFOFilter)

        self.lbCFRFiltering = QLabel(WizardDialog)
        self.lbCFRFiltering.setObjectName("lbCFRFiltering")
        self.lbCFRFiltering.setFont(font)

        self.mainLayout.addWidget(self.lbCFRFiltering)

        self.cfr_lower_bound_layout = QHBoxLayout()
        self.cfr_lower_bound_layout.setObjectName("cfr_lower_bound_layout")
        self.lbCFRLowerBound = QLabel(WizardDialog)
        self.lbCFRLowerBound.setObjectName("lbCFRLowerBound")

        self.cfr_lower_bound_layout.addWidget(self.lbCFRLowerBound)

        self.leCFRLowerBound = QLineEdit(WizardDialog)
        self.leCFRLowerBound.setObjectName("leCFRLowerBound")
        sizePolicy3.setHeightForWidth(
            self.leCFRLowerBound.sizePolicy().hasHeightForWidth()
        )
        self.leCFRLowerBound.setSizePolicy(sizePolicy3)

        self.cfr_lower_bound_layout.addWidget(self.leCFRLowerBound)

        self.mainLayout.addLayout(self.cfr_lower_bound_layout)

        self.cfr_upper_bound_layout = QHBoxLayout()
        self.cfr_upper_bound_layout.setObjectName("cfr_upper_bound_layout")
        self.lbCFRUpperBound = QLabel(WizardDialog)
        self.lbCFRUpperBound.setObjectName("lbCFRUpperBound")

        self.cfr_upper_bound_layout.addWidget(self.lbCFRUpperBound)

        self.leCFRUpperBound = QLineEdit(WizardDialog)
        self.leCFRUpperBound.setObjectName("leCFRUpperBound")
        sizePolicy3.setHeightForWidth(
            self.leCFRUpperBound.sizePolicy().hasHeightForWidth()
        )
        self.leCFRUpperBound.setSizePolicy(sizePolicy3)

        self.cfr_upper_bound_layout.addWidget(self.leCFRUpperBound)

        self.mainLayout.addLayout(self.cfr_upper_bound_layout)

        self.pbCFRFilter = QPushButton(WizardDialog)
        self.pbCFRFilter.setObjectName("pbCFRFilter")

        self.mainLayout.addWidget(self.pbCFRFilter)

        self.verticalSpacer_5 = QSpacerItem(
            20, 20, QSizePolicy.Minimum, QSizePolicy.Fixed
        )

        self.mainLayout.addItem(self.verticalSpacer_5)

        self.pbSaveData = QPushButton(WizardDialog)
        self.pbSaveData.setObjectName("pbSaveData")

        self.mainLayout.addWidget(self.pbSaveData)

        self.pbExportData = QPushButton(WizardDialog)
        self.pbExportData.setObjectName("pbExportData")

        self.mainLayout.addWidget(self.pbExportData)

        self.gridLayout.addLayout(self.mainLayout, 0, 0, 1, 1)

        QWidget.setTabOrder(self.pbLoadData, self.pbReset)
        QWidget.setTabOrder(self.pbReset, self.pbSingleColor)
        QWidget.setTabOrder(self.pbSingleColor, self.pbColorUnmixer)
        QWidget.setTabOrder(self.pbColorUnmixer, self.cmActiveColor)
        QWidget.setTabOrder(self.cmActiveColor, self.pbTimeInspector)
        QWidget.setTabOrder(self.pbTimeInspector, self.pbAnalyzer)
        QWidget.setTabOrder(self.pbAnalyzer, self.leEFOLowerBound)
        QWidget.setTabOrder(self.leEFOLowerBound, self.leEFOUpperBound)
        QWidget.setTabOrder(self.leEFOUpperBound, self.pbEFOFilter)
        QWidget.setTabOrder(self.pbEFOFilter, self.leCFRLowerBound)
        QWidget.setTabOrder(self.leCFRLowerBound, self.leCFRUpperBound)
        QWidget.setTabOrder(self.leCFRUpperBound, self.pbCFRFilter)
        QWidget.setTabOrder(self.pbCFRFilter, self.pbSaveData)
        QWidget.setTabOrder(self.pbSaveData, self.pbExportData)

        self.retranslateUi(WizardDialog)

        QMetaObject.connectSlotsByName(WizardDialog)

    # setupUi

    def retranslateUi(self, WizardDialog):
        WizardDialog.setWindowTitle(
            QCoreApplication.translate("WizardDialog", "Dialog", None)
        )
        self.pbLoadData.setText(
            QCoreApplication.translate("WizardDialog", "Load", None)
        )
        self.pbReset.setText(QCoreApplication.translate("WizardDialog", "Reset", None))
        self.lbColors.setText(
            QCoreApplication.translate("WizardDialog", "DCR unmixing", None)
        )
        self.pbSingleColor.setText(
            QCoreApplication.translate("WizardDialog", "Single fluorophore", None)
        )
        self.pbColorUnmixer.setText(
            QCoreApplication.translate("WizardDialog", "Unmixer", None)
        )
        self.lbActiveColor.setText(
            QCoreApplication.translate("WizardDialog", "Active fluorophore", None)
        )
        self.cmActiveColor.setItemText(
            0, QCoreApplication.translate("WizardDialog", "All", None)
        )

        self.lbFilters.setText(
            QCoreApplication.translate("WizardDialog", "Filters", None)
        )
        self.pbTimeInspector.setText(
            QCoreApplication.translate("WizardDialog", "Time Inspector", None)
        )
        self.pbAnalyzer.setText(
            QCoreApplication.translate("WizardDialog", "Analyzer", None)
        )
        self.lbEFOFiltering.setText(
            QCoreApplication.translate("WizardDialog", "EFO filtering", None)
        )
        self.lbEFOLowerBound.setText(
            QCoreApplication.translate("WizardDialog", "Lower bound (Hz)", None)
        )
        self.lbEFOUpperBound.setText(
            QCoreApplication.translate("WizardDialog", "Upper bound (Hz)", None)
        )
        self.pbEFOFilter.setText(
            QCoreApplication.translate("WizardDialog", "FIlter EFO", None)
        )
        self.lbCFRFiltering.setText(
            QCoreApplication.translate("WizardDialog", "CFR filtering", None)
        )
        self.lbCFRLowerBound.setText(
            QCoreApplication.translate("WizardDialog", "Lower bound", None)
        )
        self.lbCFRUpperBound.setText(
            QCoreApplication.translate("WizardDialog", "Upper bound", None)
        )
        self.pbCFRFilter.setText(
            QCoreApplication.translate("WizardDialog", "FIlter CFR", None)
        )
        self.pbSaveData.setText(
            QCoreApplication.translate("WizardDialog", "Save", None)
        )
        self.pbExportData.setText(
            QCoreApplication.translate("WizardDialog", "Export", None)
        )

    # retranslateUi



#  Copyright (c) 2022 - 2023 D-BSSE, ETH Zurich.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#   limitations under the License.
#

import numpy as np
import pyqtgraph as pg
from pyqtgraph import ROI, PlotCurveItem, PlotWidget, TextItem, ViewBox, mkPen
from PySide6 import QtCore
from PySide6.QtCore import QPoint, Qt, Signal
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMenu

from ..state import ColorCode, State
from .helpers import export_plot_interactive


class Plotter(PlotWidget):
    # Signals
    locations_selected = Signal(list, name="locations_selected")
    locations_selected_by_range = Signal(
        str, str, tuple, tuple, name="locations_selected_by_range"
    )
    crop_region_selected = Signal(str, str, tuple, tuple, name="crop_region_selected")

    def __init__(self):
        super().__init__()
        self.setBackground("w")
        self.brush = pg.mkBrush(255, 255, 255, 128)
        self.pen = pg.mkPen(None)
        self.remove_points()
        self.hideAxis("bottom")
        self.hideAxis("left")
        self.show()

        # Disable default context menu
        self.setMenuEnabled(False)

        # Set aspect ratio to 1.0 locked
        self.getPlotItem().getViewBox().setAspectLocked(lock=True, ratio=1.0)

        # Keep a reference to the singleton State class
        self.state = State()

        # Keep a reference to the scatter plot object
        self.scatter = None

        # ROI for localizations selection
        self.ROI = None
        self.line = None
        self.line_text = None
        self._roi_start_point = None
        self._roi_is_being_drawn = False
        self._line_start_point = None
        self._line_is_being_drawn = False

        # Keep track of last plot
        self._last_x_param = None
        self._last_y_param = None

    def enableAutoRange(self, enable: bool):
        """Enable/disable axes autorange."""
        self.getViewBox().enableAutoRange(axis=ViewBox.XYAxes, enable=enable)

    def mousePressEvent(self, ev):
        """Override mouse press event."""

        # Is the user trying to initiate drawing an ROI?
        if (
            self.scatter is not None
            and ev.button() == Qt.MouseButton.LeftButton
            and ev.modifiers() == QtCore.Qt.ShiftModifier
        ):
            # Remove previous ROI if it exists
            if self.ROI is not None:
                self.removeItem(self.ROI)
                self.ROI.deleteLater()
                self.ROI = None

            # Create ROI and keep track of position
            self._roi_is_being_drawn = True
            self._roi_start_point = (
                self.getPlotItem().getViewBox().mapSceneToView(ev.position())
            )
            self.ROI = ROI(
                pos=self._roi_start_point,
                size=(0, 0),
                resizable=False,
                rotatable=False,
                pen=(255, 0, 0),
            )
            self.ROI.setAcceptedMouseButtons(Qt.MouseButton.RightButton)
            self.addItem(self.ROI)
            self.ROI.show()

            # Make sure to react to ROI shifts
            self.ROI.sigRegionChangeFinished.connect(self.roi_moved)

            # Accept the event
            ev.accept()

        elif (
            self.scatter is not None
            and ev.button() == Qt.MouseButton.LeftButton
            and ev.modifiers() == QtCore.Qt.ControlModifier
            and self.state.x_param in ("x", "y")
            and self.state.y_param in ("x", "y")
        ):
            # Is the user trying to initiate drawing a line?

            # Remove previous line if it exists
            if self.line is not None:
                self.removeItem(self.line)
                self.line.deleteLater()
                self.line = None

            if self.line_text is not None:
                self.removeItem(self.line_text)
                self.line_text.deleteLater()
                self.line_text = None

            # Create ROI and keep track of position
            self._line_is_being_drawn = True
            self._line_start_point = (
                self.getPlotItem().getViewBox().mapSceneToView(ev.position())
            )
            self.line = PlotCurveItem(
                x=[self._line_start_point.x(), self._line_start_point.x()],
                y=[self._line_start_point.y(), self._line_start_point.y()],
                pen=mkPen(color=(255, 0, 0), width=2),
            )
            self.addItem(self.line)
            self.line.show()

            # Accept the event
            ev.accept()

        else:

            # Is the user trying to open a context menu?
            if self.scatter is not None and ev.button() == Qt.MouseButton.RightButton:
                menu = QMenu()
                if self.ROI is not None:
                    crop_data_action = QAction("Crop data")
                    crop_data_action.triggered.connect(self.crop_data_by_roi_selection)
                    menu.addAction(crop_data_action)
                    menu.addSeparator()
                export_action = QAction("Export plot")
                export_action.triggered.connect(
                    lambda checked: export_plot_interactive(self.getPlotItem())
                )
                menu.addAction(export_action)
                pos = ev.screenPos()
                menu.exec(QPoint(int(pos.x()), int(pos.y())))
                ev.accept()
            else:

                # Call the parent method
                ev.ignore()
                super().mousePressEvent(ev)

    def mouseMoveEvent(self, ev):
        # Is the user drawing an ROI?
        if (
            self.scatter is not None
            and ev.buttons() == Qt.MouseButton.LeftButton
            and self._roi_is_being_drawn
        ):
            # Resize the ROI
            current_point = (
                self.getPlotItem().getViewBox().mapSceneToView(ev.position())
            )
            self.ROI.setSize(current_point - self._roi_start_point)

            # Accept the event
            ev.accept()

        elif (
            self.scatter is not None
            and ev.buttons() == Qt.MouseButton.LeftButton
            and self._line_is_being_drawn
        ):
            # Is the user drawing a line?

            # Resize the ROI
            current_point = (
                self.getPlotItem().getViewBox().mapSceneToView(ev.position())
            )
            self.line.setData(
                x=[self._line_start_point.x(), current_point.x()],
                y=[self._line_start_point.y(), current_point.y()],
            )

            # Accept the event
            ev.accept()
        else:
            # Call the parent method
            ev.ignore()
            super().mouseMoveEvent(ev)

    def mouseReleaseEvent(self, ev):
        if (
            self.scatter is not None
            and ev.button() == Qt.MouseButton.LeftButton
            and self._roi_is_being_drawn
        ):
            # Extract the ranges
            x_range, y_range = self._get_ranges_from_roi()

            # Get current parameters
            x_param = self.state.x_param
            y_param = self.state.y_param

            # Update the DataViewer with current selection
            if x_range is not None and y_range is not None:
                self.locations_selected_by_range.emit(
                    x_param, y_param, x_range, y_range
                )

            # Reset flags
            self._roi_start_point = None
            self._roi_is_being_drawn = False

            # If the ROI has 0 size (when a shift-click has been used to remove a previous ROI),
            # clean it up properly.
            if (
                np.abs(x_range[1] - x_range[0]) < 1e-4
                or np.abs(y_range[1] - y_range[0]) < 1e-4
            ):
                self.removeItem(self.ROI)
                self.ROI.deleteLater()
                self.ROI = None

            # Accept the event
            ev.accept()

        elif (
            self.scatter is not None
            and ev.button() == Qt.MouseButton.LeftButton
            and self._line_is_being_drawn
        ):
            # Display the measurement
            x_data, y_data = self.line.getData()
            center = np.array(
                [0.5 * (x_data[0] + x_data[1]), 0.5 * (y_data[0] + y_data[1])]
            )
            delta_y = np.array(y_data[1] - y_data[0])
            delta_x = np.array(x_data[1] - x_data[0])
            v = np.array([-1.0 * delta_y, delta_x])
            norm = np.linalg.norm(v)
            if np.abs(norm) >= 1e-4:
                v_norm = v / norm
                length = np.sqrt(delta_x * delta_x + delta_y * delta_y)
                pos = center + 1.0 * v_norm
                if self.state.color_code == ColorCode.NONE:
                    clr = (255, 255, 0)
                else:
                    clr = (255, 255, 255)
                self.line_text = TextItem(text=f"{length:.2f} nm", color=clr)
                self.line_text.setPos(pos[0], pos[1])
                self.addItem(self.line_text)
            else:
                if self.line is not None:
                    self.removeItem(self.line)
                    self.line.deleteLater()
                    self.line = None

            # Reset flags
            self._line_start_point = None
            self._line_is_being_drawn = False

            # Accept the event
            ev.accept()

        else:
            # Call the parent method
            ev.ignore()
            super().mouseReleaseEvent(ev)

    def remove_points(self):
        self.setBackground("w")
        self.clear()

    def plot_parameters(self, x, y, x_param, y_param, tid, fid):
        """Plot localizations in a 2D scatter plot."""

        # Create the scatter plot
        self.scatter = pg.ScatterPlotItem(
            size=5,
            pen=self.pen,
            brush=self.brush,
            hoverable=True,
            hoverSymbol="s",
            hoverSize=5,
            hoverPen=pg.mkPen("w", width=2),
            hoverBrush=None,
        )
        self.scatter.sigClicked.connect(self.clicked)
        if self.state.color_code == ColorCode.NONE:
            brushes = self.brush
        elif self.state.color_code == ColorCode.BY_TID:
            brushes = tid
        elif self.state.color_code == ColorCode.BY_FLUO:
            brushes = [6 if i == 1 else 9 for i in fid]
        else:
            raise ValueError("Unexpected request for color-coding the localizations!")

        self.scatter.addPoints(
            x=x,
            y=y,
            data=tid,
            brush=brushes,
        )
        self.addItem(self.scatter)
        self.setLabel("bottom", text=x_param)
        self.setLabel("left", text=y_param)
        self.showAxis("bottom")
        self.showAxis("left")
        self.setBackground("k")

        if (self._last_x_param is None or self._last_x_param != x_param) or (
            self._last_y_param is None or self._last_y_param != y_param
        ):
            # Update range
            self.getViewBox().enableAutoRange(axis=ViewBox.XYAxes, enable=True)

            # Fix aspect ratio
            x_scale = (x.max() - x.min()) / (len(x) - 1)
            y_scale = (y.max() - y.min()) / (len(y) - 1)
            aspect_ratio = y_scale / x_scale
            self.getPlotItem().getViewBox().setAspectLocked(
                lock=True, ratio=aspect_ratio
            )

        # Update last plotted parameters
        self._last_x_param = x_param
        self._last_y_param = y_param

    def crop_data_by_roi_selection(self, item):
        """Open dialog to manually set the filter ranges"""

        # Get the ROI bounds
        pos = self.ROI.pos()
        size = self.ROI.size()

        # Create the ranges
        x_range = (pos[0], pos[0] + size[0])
        y_range = (pos[1], pos[1] + size[1])

        # Get the parameter names
        x_param = self.state.x_param
        y_param = self.state.y_param

        self.crop_region_selected.emit(x_param, y_param, x_range, y_range)

        # Delete ROI after it has been used to crop
        self.removeItem(self.ROI)
        self.ROI.deleteLater()
        self.ROI = None

    def roi_moved(self):
        """Inform that the selection of localizations may have changed after the ROI was moved."""

        # If the ROI is being drawn now, do nothing
        if self._roi_is_being_drawn:
            return

        # Extract the ranges
        x_range, y_range = self._get_ranges_from_roi()

        # Get current parameters
        x_param = self.state.x_param
        y_param = self.state.y_param

        # Update the DataViewer with current selection
        if x_range is not None and y_range is not None:
            self.locations_selected_by_range.emit(x_param, y_param, x_range, y_range)

    def clicked(self, _, points):
        """Emit 'signal_selected_locations' when points are selected in the plot."""
        self.locations_selected.emit(points)

        # Remove ROI if it exists
        if self.ROI is not None:
            self.removeItem(self.ROI)
            self.ROI = None

    def _get_ranges_from_roi(self):
        """Calculate x and y ranges from ROI."""

        # Initialize x and y ranges to None
        x_range = None
        y_range = None

        # Extract x and y ranges
        if self.ROI is not None:
            x_range = (self.ROI.pos()[0], self.ROI.pos()[0] + self.ROI.size()[0])
            y_range = (self.ROI.pos()[1], self.ROI.pos()[1] + self.ROI.size()[1])

        return x_range, y_range



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'frc_tool.ui'
##
## Created by: Qt User Interface Compiler version 6.4.3
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (
    QCoreApplication,
    QDate,
    QDateTime,
    QLocale,
    QMetaObject,
    QObject,
    QPoint,
    QRect,
    QSize,
    Qt,
    QTime,
    QUrl,
)
from PySide6.QtGui import (
    QBrush,
    QColor,
    QConicalGradient,
    QCursor,
    QFont,
    QFontDatabase,
    QGradient,
    QIcon,
    QImage,
    QKeySequence,
    QLinearGradient,
    QPainter,
    QPalette,
    QPixmap,
    QRadialGradient,
    QTransform,
)
from PySide6.QtWidgets import (
    QApplication,
    QCheckBox,
    QDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QLineEdit,
    QProgressBar,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QWidget,
)


class Ui_FRCTool(object):
    def setupUi(self, FRCTool):
        if not FRCTool.objectName():
            FRCTool.setObjectName("FRCTool")
        FRCTool.resize(1050, 800)
        FRCTool.setMinimumSize(QSize(800, 600))
        self.gridLayout = QGridLayout(FRCTool)
        self.gridLayout.setObjectName("gridLayout")
        self.hlParameters = QHBoxLayout()
        self.hlParameters.setObjectName("hlParameters")
        self.hlParameters.setSizeConstraint(QLayout.SetFixedSize)
        self.hsCFRFilterBefore = QSpacerItem(
            20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlParameters.addItem(self.hsCFRFilterBefore)

        self.lbLateralResolution = QLabel(FRCTool)
        self.lbLateralResolution.setObjectName("lbLateralResolution")
        sizePolicy = QSizePolicy(QSizePolicy.Preferred, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.lbLateralResolution.sizePolicy().hasHeightForWidth()
        )
        self.lbLateralResolution.setSizePolicy(sizePolicy)

        self.hlParameters.addWidget(self.lbLateralResolution)

        self.leLateralResolution = QLineEdit(FRCTool)
        self.leLateralResolution.setObjectName("leLateralResolution")
        sizePolicy.setHeightForWidth(
            self.leLateralResolution.sizePolicy().hasHeightForWidth()
        )
        self.leLateralResolution.setSizePolicy(sizePolicy)

        self.hlParameters.addWidget(self.leLateralResolution)

        self.cbEndpoint = QCheckBox(FRCTool)
        self.cbEndpoint.setObjectName("cbEndpoint")

        self.hlParameters.addWidget(self.cbEndpoint)

        self.lbTemporalResolution = QLabel(FRCTool)
        self.lbTemporalResolution.setObjectName("lbTemporalResolution")
        sizePolicy.setHeightForWidth(
            self.lbTemporalResolution.sizePolicy().hasHeightForWidth()
        )
        self.lbTemporalResolution.setSizePolicy(sizePolicy)

        self.hlParameters.addWidget(self.lbTemporalResolution)

        self.leTemporalResolution = QLineEdit(FRCTool)
        self.leTemporalResolution.setObjectName("leTemporalResolution")

        self.hlParameters.addWidget(self.leTemporalResolution)

        self.lbNumRepeats = QLabel(FRCTool)
        self.lbNumRepeats.setObjectName("lbNumRepeats")
        sizePolicy.setHeightForWidth(self.lbNumRepeats.sizePolicy().hasHeightForWidth())
        self.lbNumRepeats.setSizePolicy(sizePolicy)

        self.hlParameters.addWidget(self.lbNumRepeats)

        self.leNumRepeats = QLineEdit(FRCTool)
        self.leNumRepeats.setObjectName("leNumRepeats")
        sizePolicy.setHeightForWidth(self.leNumRepeats.sizePolicy().hasHeightForWidth())
        self.leNumRepeats.setSizePolicy(sizePolicy)

        self.hlParameters.addWidget(self.leNumRepeats)

        self.pbRunFRCAnalysis = QPushButton(FRCTool)
        self.pbRunFRCAnalysis.setObjectName("pbRunFRCAnalysis")
        sizePolicy.setHeightForWidth(
            self.pbRunFRCAnalysis.sizePolicy().hasHeightForWidth()
        )
        self.pbRunFRCAnalysis.setSizePolicy(sizePolicy)

        self.hlParameters.addWidget(self.pbRunFRCAnalysis)

        self.hsCFRFilterAfter = QSpacerItem(
            20, 20, QSizePolicy.Expanding, QSizePolicy.Minimum
        )

        self.hlParameters.addItem(self.hsCFRFilterAfter)

        self.gridLayout.addLayout(self.hlParameters, 0, 0, 1, 1)

        self.hlPlot = QHBoxLayout()
        self.hlPlot.setObjectName("hlPlot")

        self.gridLayout.addLayout(self.hlPlot, 1, 0, 1, 1)

        self.hlProgress = QHBoxLayout()
        self.hlProgress.setObjectName("hlProgress")
        self.progress_bar = QProgressBar(FRCTool)
        self.progress_bar.setObjectName("progress_bar")
        self.progress_bar.setValue(0)

        self.hlProgress.addWidget(self.progress_bar)

        self.pbInterrupt = QPushButton(FRCTool)
        self.pbInterrupt.setObjectName("pbInterrupt")
        sizePolicy1 = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.pbInterrupt.sizePolicy().hasHeightForWidth())
        self.pbInterrupt.setSizePolicy(sizePolicy1)

        self.hlProgress.addWidget(self.pbInterrupt)

        self.gridLayout.addLayout(self.hlProgress, 2, 0, 1, 1)

        self.retranslateUi(FRCTool)

        QMetaObject.connectSlotsByName(FRCTool)

    # setupUi

    def retranslateUi(self, FRCTool):
        FRCTool.setWindowTitle(
            QCoreApplication.translate("FRCTool", "FRC Analyzer", None)
        )
        self.lbLateralResolution.setText(
            QCoreApplication.translate("FRCTool", "Spatial (xy) resolution (nm)", None)
        )
        self.cbEndpoint.setText(
            QCoreApplication.translate("FRCTool", "Endpoint estimation", None)
        )
        self.lbTemporalResolution.setText(
            QCoreApplication.translate("FRCTool", "Temporal resoluton (s)", None)
        )
        self.lbNumRepeats.setText(
            QCoreApplication.translate("FRCTool", "Number of repeats", None)
        )
        self.pbRunFRCAnalysis.setText(
            QCoreApplication.translate("FRCTool", "Run", None)
        )
        self.pbInterrupt.setText(QCoreApplication.translate("FRCTool", "Cancel", None))

    # retranslateUi
